# encoding: utf-8
import logging
import sys
from threading import Thread, Condition

from ._MbContextMessageBus import MbContextMessageBus
from messagebus import Event, Message, DataType, TokenCreator
from messagehandler import MessageHandlerBuilder
from msgbus import (
    SE_SUCCESS,
    TK_CMP_TERM_NOW,
    TK_CMP_START,
    TK_EVT_EVENT,
    TK_SYS_ACK,
    TK_SYS_NAK,
    TK_HV_SERVICEONLINE,
    TK_HV_SERVICEOFFLINE,
)
from systools import sys_log_error
from typing import Dict, List, Optional

logger = logging.getLogger(__name__)


class MbContextMessageHandler(object):
    def __init__(self, message_bus, service_name, service_type, required_services, event_handler_builder):
        # type: (MbContextMessageBus, str, str, str, MessageHandlerBuilder) -> None
        self.message_bus = message_bus
        self.service_name = service_name
        self.service_type = service_type
        self.required_services = required_services
        self.event_handler_builder = event_handler_builder
        self.reentrant_subjects = {}
        self.non_reentrant_subjects = {}
        self.thread_handling_subject = {}

        self.queue_subjects = {}
        self.event_queue = []  # type: List[Event]
        self.event_queue_thread = None  # type: Thread
        self.event_queue_condition = Condition()  # type: Condition

        self.queue_tokens = {}
        self.message_queue = []  # type: List[Message]
        self.message_queue_thread = None  # type: Thread
        self.message_queue_condition = Condition()  # type: Condition

        self.group_to_filter = None  # type: Optional[str]
        self.finished = False

        if self.service_name and self.service_type != "":
            status = self.message_bus.mbcontext.MB_EasyWaitStart(
                "{}:{}".format(self.service_name, self.service_type),
                self.required_services
            )
            if status != SE_SUCCESS:
                sys_log_error("Terminating component after 'MB_EasyWaitStart'")
                sys.exit(status)

    def subscribe_reentrant_events(self, subjects):
        # type: (List[unicode]) -> MbContextMessageHandler
        for subject in subjects:
            self.reentrant_subjects[subject] = subject
            self.message_bus.subscribe(subject)

        return self

    def subscribe_non_reentrant_events(self, subjects):
        # type: (List[unicode]) -> MbContextMessageHandler
        for subject in subjects:
            self.non_reentrant_subjects[subject] = subject
            self.message_bus.subscribe(subject)

        return self

    def subscribe_queue_events(self, subjects):
        # type: (List[unicode]) -> MbContextMessageHandler

        subject_event_queue_thread = Thread(
            target=self._event_queue_method,
            name="Event Queue Thread",
        )
        subject_event_queue_thread.daemon = True
        subject_event_queue_thread.start()
        self.queue_subjects = subjects

        for subject in subjects:
            self.message_bus.subscribe(subject)

        return self

    def subscribe_queue_messages(self, tokens):
        # type: (List[int]) -> MbContextMessageHandler

        token_message_queue_thread = Thread(
            target=self._message_queue_method,
            name="Message Queue Thread",
        )
        token_message_queue_thread.daemon = True
        token_message_queue_thread.start()
        self.queue_tokens = tokens

        return self

    def handle_events(self):
        self.event_handler_builder.build_singletons()

        while not self.finished:
            msg = self.message_bus.mbcontext.MB_EasyGetMessage()

            # Check for component terminate request
            if (not msg) or (msg.token == TK_CMP_TERM_NOW):
                # Component terminated, warn the handler that we are terminated
                self.event_handler_builder.destroy_singletons()
                # reply the message
                self.message_bus.mbcontext.MB_EasyReplyMessage(msg)
                # Finish the loop and terminate the component
                self.finished = True
                break

            if msg.token in (TK_CMP_START, TK_HV_SERVICEONLINE, TK_HV_SERVICEOFFLINE):
                msg.token = TK_SYS_ACK
                self.message_bus.mbcontext.MB_EasyReplyMessage(msg)

            elif msg.token == TK_EVT_EVENT:
                params = msg.data.split(b'\x00')
                data = params[0] if len(params) > 0 else None
                subject = params[1] if len(params) > 1 else None
                subject = subject.decode("utf-8")
                evt_type = params[2] if len(params) > 2 else None
                if evt_type == b'':
                    evt_type = None
                if evt_type is not None:
                    evt_type = evt_type.decode("utf-8")
                sync = params[3] == b'true' if len(params) > 3 else False
                source_id = int(params[4]) if len(params) > 4 else None

                event = Event(
                    subject=subject,
                    event_type=evt_type,
                    data=data,
                    source=source_id,
                    synchronous=sync,
                    headers=self.parse_headers(msg.headers),
                    ttl=msg.ttl,
                )
                event.imp_message = msg

                if subject in self.non_reentrant_subjects:
                    handling_thread = False
                    if subject in self.thread_handling_subject:
                        handling_thread = self.thread_handling_subject[subject].is_alive()

                    # Se o evento é não reentrante e já temos uma thread tratando outra instância do mesmo evento,
                    # respondemos e não criamos outra thread
                    if handling_thread:
                        self.message_bus.mbcontext.MB_EasyReplyMessage(msg)
                        continue

                if subject in self.queue_subjects:
                    # Se o evento é do tipo fila, enfileiramos a tupla e deixamos a thread responsável tratá-lo
                    with self.event_queue_condition:
                        self.event_queue.append(event)
                        self.event_queue_condition.notify()

                    self.message_bus.mbcontext.MB_EasyReplyMessage(msg)
                    continue

                handler_thread = Thread(target=self._handle_event, args=(event,))
                handler_thread.daemon = True

                if subject in self.non_reentrant_subjects:
                    self.thread_handling_subject[subject] = handler_thread

                handler_thread.start()

                # Se o evento for syncrono, deixamos a thread responsável por dar o reply na mensagem
                if not sync:
                    self.message_bus.mbcontext.MB_EasyReplyMessage(msg)
            else:
                token = msg.token
                if self.group_to_filter is not None and TokenCreator.is_in_group(token, self.group_to_filter):
                    msg.token = TK_SYS_NAK
                    self.message_bus.mbcontext.MB_EasyReplyMessage(msg)
                    continue

                data = msg.data
                data_format = msg.format
                if msg.data is None or msg.data == "":
                    data = None
                    data_format = DataType.empty

                message = Message(
                    token=token,
                    data_type=data_format,
                    data=data,
                    headers=self.parse_headers(msg.headers),
                    ttl=msg.ttl,
                )
                message.imp_message = msg

                if token in self.queue_tokens:
                    # Se a mensagem é do tipo fila, enfileiramos a tupla e deixamos a thread responsável tratá-lo
                    with self.message_queue_condition:
                        self.message_queue.append(message)
                        self.message_queue_condition.notify()

                    continue

                handler_thread = Thread(target=self._handle_message, args=(message,))
                handler_thread.daemon = True
                handler_thread.start()

    def _message_queue_method(self):
        my_condition = self.message_queue_condition
        my_queue = self.message_queue

        self._handle_each_queued_object(my_condition, my_queue, self._handle_message)

    def _event_queue_method(self):
        my_condition = self.event_queue_condition
        my_queue = self.event_queue

        self._handle_each_queued_object(my_condition, my_queue, self._handle_event)

    def _handle_each_queued_object(self, my_condition, my_queue, handler_fn):
        while not self.finished:
            while len(my_queue) > 0:
                with my_condition:
                    obj = my_queue.pop(0)

                try:
                    handler_fn(obj)
                except Exception: # noqa
                    pass

            with my_condition:
                my_condition.wait()

    def _handle_event(self, event):
        # type: (Event) -> None
        event_handler = None
        try:
            event_handler = self.event_handler_builder.build_message_handler(None, event)
            event_handler.handle_event(self.message_bus, event)
            if event.synchronous and not hasattr(event, '_replied'):
                logger.error("Sync event without reply: {}".format(event.subject))
        finally:
            if event_handler is not None:
                self.event_handler_builder.destroy_message_handler(event_handler)

    def _handle_message(self, message):
        # type: (Message) -> None
        message_handler = None
        try:
            message_handler = self.event_handler_builder.build_message_handler(message, None)
            message_handler.handle_message(self.message_bus, message)
        finally:
            if message_handler is not None:
                self.event_handler_builder.destroy_message_handler(message_handler)

    @staticmethod
    def parse_headers(
        headers,         # type: bytes
    ):
        if headers is None or headers == b"":
            return {}

        parts = headers.split(b"\0")
        header_qty = int(parts[0])
        headers = {}
        for i in range(header_qty):
            headers[parts[1 + 2 * i].decode("utf-8")] = parts[2 + 2 * i].decode("utf-8")
        return headers



from production.box import NotShowOnKitchenAddTagBox
from production.treebuilder import BoxConfiguration

from ._BoxBuilder import BoxBuilder


class NotShowOnKitchenAddTagBoxBuilder(BoxBuilder):
    def __init__(self, product_repository, loggers):
        super(NotShowOnKitchenAddTagBoxBuilder, self).__init__(loggers)

        self.product_repository = product_repository

    def build(self, config):
        # type: (BoxConfiguration) -> NotShowOnKitchenAddTagBox

        tag_to_add = self.get_extra(config, "TagToAdd", None)

        return NotShowOnKitchenAddTagBox(
            name=config.name,
            sons=config.sons,
            tag_to_add=tag_to_add,
            product_repository=self.product_repository,
            logger=self.get_logger(config),
        )



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=too-many-ancestors

"""
The OpenTelemetry metrics API  describes the classes used to generate
metrics.

The :class:`.MeterProvider` provides users access to the :class:`.Meter` which in
turn is used to create :class:`.Instrument` objects. The :class:`.Instrument` objects are
used to record measurements.

This module provides abstract (i.e. unimplemented) classes required for
metrics, and a concrete no-op implementation :class:`.NoOpMeter` that allows applications
to use the API package alone without a supporting implementation.

To get a meter, you need to provide the package name from which you are
calling the meter APIs to OpenTelemetry by calling `MeterProvider.get_meter`
with the calling instrumentation name and the version of your package.

The following code shows how to obtain a meter using the global :class:`.MeterProvider`::

    from opentelemetry.metrics import get_meter

    meter = get_meter("example-meter")
    counter = meter.create_counter("example-counter")

.. versionadded:: 1.10.0
"""


from abc import ABC, abstractmethod
from logging import getLogger
from os import environ
from threading import Lock
from typing import List, Optional, Sequence, Set, Tuple, Union, cast

from opentelemetry.environment_variables import OTEL_PYTHON_METER_PROVIDER
from opentelemetry.metrics._internal.instrument import (
    CallbackT,
    Counter,
    Histogram,
    NoOpCounter,
    NoOpHistogram,
    NoOpObservableCounter,
    NoOpObservableGauge,
    NoOpObservableUpDownCounter,
    NoOpUpDownCounter,
    ObservableCounter,
    ObservableGauge,
    ObservableUpDownCounter,
    UpDownCounter,
    _ProxyCounter,
    _ProxyHistogram,
    _ProxyObservableCounter,
    _ProxyObservableGauge,
    _ProxyObservableUpDownCounter,
    _ProxyUpDownCounter,
)
from opentelemetry.util._once import Once
from opentelemetry.util._providers import _load_provider

_logger = getLogger(__name__)


_ProxyInstrumentT = Union[
    _ProxyCounter,
    _ProxyHistogram,
    _ProxyObservableCounter,
    _ProxyObservableGauge,
    _ProxyObservableUpDownCounter,
    _ProxyUpDownCounter,
]


class MeterProvider(ABC):
    """
    MeterProvider is the entry point of the API. It provides access to `Meter` instances.
    """

    @abstractmethod
    def get_meter(
        self,
        name: str,
        version: Optional[str] = None,
        schema_url: Optional[str] = None,
    ) -> "Meter":
        """Returns a `Meter` for use by the given instrumentation library.

        For any two calls it is undefined whether the same or different
        `Meter` instances are returned, even for different library names.

        This function may return different `Meter` types (e.g. a no-op meter
        vs. a functional meter).

        Args:
            name: The name of the instrumenting module.
                ``__name__`` may not be used as this can result in
                different meter names if the meters are in different files.
                It is better to use a fixed string that can be imported where
                needed and used consistently as the name of the meter.

                This should *not* be the name of the module that is
                instrumented but the name of the module doing the instrumentation.
                E.g., instead of ``"requests"``, use
                ``"opentelemetry.instrumentation.requests"``.

            version: Optional. The version string of the
                instrumenting library.  Usually this should be the same as
                ``importlib.metadata.version(instrumenting_library_name)``.

            schema_url: Optional. Specifies the Schema URL of the emitted telemetry.
        """


class NoOpMeterProvider(MeterProvider):
    """The default MeterProvider used when no MeterProvider implementation is available."""

    def get_meter(
        self,
        name: str,
        version: Optional[str] = None,
        schema_url: Optional[str] = None,
    ) -> "Meter":
        """Returns a NoOpMeter."""
        super().get_meter(name, version=version, schema_url=schema_url)
        return NoOpMeter(name, version=version, schema_url=schema_url)


class _ProxyMeterProvider(MeterProvider):
    def __init__(self) -> None:
        self._lock = Lock()
        self._meters: List[_ProxyMeter] = []
        self._real_meter_provider: Optional[MeterProvider] = None

    def get_meter(
        self,
        name: str,
        version: Optional[str] = None,
        schema_url: Optional[str] = None,
    ) -> "Meter":
        with self._lock:
            if self._real_meter_provider is not None:
                return self._real_meter_provider.get_meter(
                    name, version, schema_url
                )

            meter = _ProxyMeter(name, version=version, schema_url=schema_url)
            self._meters.append(meter)
            return meter

    def on_set_meter_provider(self, meter_provider: MeterProvider) -> None:
        with self._lock:
            self._real_meter_provider = meter_provider
            for meter in self._meters:
                meter.on_set_meter_provider(meter_provider)


class Meter(ABC):
    """Handles instrument creation.

    This class provides methods for creating instruments which are then
    used to produce measurements.
    """

    def __init__(
        self,
        name: str,
        version: Optional[str] = None,
        schema_url: Optional[str] = None,
    ) -> None:
        super().__init__()
        self._name = name
        self._version = version
        self._schema_url = schema_url
        self._instrument_ids: Set[str] = set()
        self._instrument_ids_lock = Lock()

    @property
    def name(self) -> str:
        """
        The name of the instrumenting module.
        """
        return self._name

    @property
    def version(self) -> Optional[str]:
        """
        The version string of the instrumenting library.
        """
        return self._version

    @property
    def schema_url(self) -> Optional[str]:
        """
        Specifies the Schema URL of the emitted telemetry
        """
        return self._schema_url

    def _is_instrument_registered(
        self, name: str, type_: type, unit: str, description: str
    ) -> Tuple[bool, str]:
        """
        Check if an instrument with the same name, type, unit and description
        has been registered already.

        Returns a tuple. The first value is `True` if the instrument has been
        registered already, `False` otherwise. The second value is the
        instrument id.
        """

        instrument_id = ",".join(
            [name.strip().lower(), type_.__name__, unit, description]
        )

        result = False

        with self._instrument_ids_lock:
            if instrument_id in self._instrument_ids:
                result = True
            else:
                self._instrument_ids.add(instrument_id)

        return (result, instrument_id)

    @abstractmethod
    def create_counter(
        self,
        name: str,
        unit: str = "",
        description: str = "",
    ) -> Counter:
        """Creates a `Counter` instrument

        Args:
            name: The name of the instrument to be created
            unit: The unit for observations this instrument reports. For
                example, ``By`` for bytes. UCUM units are recommended.
            description: A description for this instrument and what it measures.
        """

    @abstractmethod
    def create_up_down_counter(
        self,
        name: str,
        unit: str = "",
        description: str = "",
    ) -> UpDownCounter:
        """Creates an `UpDownCounter` instrument

        Args:
            name: The name of the instrument to be created
            unit: The unit for observations this instrument reports. For
                example, ``By`` for bytes. UCUM units are recommended.
            description: A description for this instrument and what it measures.
        """

    @abstractmethod
    def create_observable_counter(
        self,
        name: str,
        callbacks: Optional[Sequence[CallbackT]] = None,
        unit: str = "",
        description: str = "",
    ) -> ObservableCounter:
        """Creates an `ObservableCounter` instrument

        An observable counter observes a monotonically increasing count by calling provided
        callbacks which accept a :class:`~opentelemetry.metrics.CallbackOptions` and return
        multiple :class:`~opentelemetry.metrics.Observation`.

        For example, an observable counter could be used to report system CPU
        time periodically. Here is a basic implementation::

            def cpu_time_callback(options: CallbackOptions) -> Iterable[Observation]:
                observations = []
                with open("/proc/stat") as procstat:
                    procstat.readline()  # skip the first line
                    for line in procstat:
                        if not line.startswith("cpu"): break
                        cpu, *states = line.split()
                        observations.append(Observation(int(states[0]) // 100, {"cpu": cpu, "state": "user"}))
                        observations.append(Observation(int(states[1]) // 100, {"cpu": cpu, "state": "nice"}))
                        observations.append(Observation(int(states[2]) // 100, {"cpu": cpu, "state": "system"}))
                        # ... other states
                return observations

            meter.create_observable_counter(
                "system.cpu.time",
                callbacks=[cpu_time_callback],
                unit="s",
                description="CPU time"
            )

        To reduce memory usage, you can use generator callbacks instead of
        building the full list::

            def cpu_time_callback(options: CallbackOptions) -> Iterable[Observation]:
                with open("/proc/stat") as procstat:
                    procstat.readline()  # skip the first line
                    for line in procstat:
                        if not line.startswith("cpu"): break
                        cpu, *states = line.split()
                        yield Observation(int(states[0]) // 100, {"cpu": cpu, "state": "user"})
                        yield Observation(int(states[1]) // 100, {"cpu": cpu, "state": "nice"})
                        # ... other states

        Alternatively, you can pass a sequence of generators directly instead of a sequence of
        callbacks, which each should return iterables of :class:`~opentelemetry.metrics.Observation`::

            def cpu_time_callback(states_to_include: set[str]) -> Iterable[Iterable[Observation]]:
                # accept options sent in from OpenTelemetry
                options = yield
                while True:
                    observations = []
                    with open("/proc/stat") as procstat:
                        procstat.readline()  # skip the first line
                        for line in procstat:
                            if not line.startswith("cpu"): break
                            cpu, *states = line.split()
                            if "user" in states_to_include:
                                observations.append(Observation(int(states[0]) // 100, {"cpu": cpu, "state": "user"}))
                            if "nice" in states_to_include:
                                observations.append(Observation(int(states[1]) // 100, {"cpu": cpu, "state": "nice"}))
                            # ... other states
                    # yield the observations and receive the options for next iteration
                    options = yield observations

            meter.create_observable_counter(
                "system.cpu.time",
                callbacks=[cpu_time_callback({"user", "system"})],
                unit="s",
                description="CPU time"
            )

        The :class:`~opentelemetry.metrics.CallbackOptions` contain a timeout which the
        callback should respect. For example if the callback does asynchronous work, like
        making HTTP requests, it should respect the timeout::

            def scrape_http_callback(options: CallbackOptions) -> Iterable[Observation]:
                r = requests.get('http://scrapethis.com', timeout=options.timeout_millis / 10**3)
                for value in r.json():
                    yield Observation(value)

        Args:
            name: The name of the instrument to be created
            callbacks: A sequence of callbacks that return an iterable of
                :class:`~opentelemetry.metrics.Observation`. Alternatively, can be a sequence of generators that each
                yields iterables of :class:`~opentelemetry.metrics.Observation`.
            unit: The unit for observations this instrument reports. For
                example, ``By`` for bytes. UCUM units are recommended.
            description: A description for this instrument and what it measures.
        """

    @abstractmethod
    def create_histogram(
        self,
        name: str,
        unit: str = "",
        description: str = "",
    ) -> Histogram:
        """Creates a :class:`~opentelemetry.metrics.Histogram` instrument

        Args:
            name: The name of the instrument to be created
            unit: The unit for observations this instrument reports. For
                example, ``By`` for bytes. UCUM units are recommended.
            description: A description for this instrument and what it measures.
        """

    @abstractmethod
    def create_observable_gauge(
        self,
        name: str,
        callbacks: Optional[Sequence[CallbackT]] = None,
        unit: str = "",
        description: str = "",
    ) -> ObservableGauge:
        """Creates an `ObservableGauge` instrument

        Args:
            name: The name of the instrument to be created
            callbacks: A sequence of callbacks that return an iterable of
                :class:`~opentelemetry.metrics.Observation`. Alternatively, can be a generator that yields iterables
                of :class:`~opentelemetry.metrics.Observation`.
            unit: The unit for observations this instrument reports. For
                example, ``By`` for bytes. UCUM units are recommended.
            description: A description for this instrument and what it measures.
        """

    @abstractmethod
    def create_observable_up_down_counter(
        self,
        name: str,
        callbacks: Optional[Sequence[CallbackT]] = None,
        unit: str = "",
        description: str = "",
    ) -> ObservableUpDownCounter:
        """Creates an `ObservableUpDownCounter` instrument

        Args:
            name: The name of the instrument to be created
            callbacks: A sequence of callbacks that return an iterable of
                :class:`~opentelemetry.metrics.Observation`. Alternatively, can be a generator that yields iterables
                of :class:`~opentelemetry.metrics.Observation`.
            unit: The unit for observations this instrument reports. For
                example, ``By`` for bytes. UCUM units are recommended.
            description: A description for this instrument and what it measures.
        """


class _ProxyMeter(Meter):
    def __init__(
        self,
        name: str,
        version: Optional[str] = None,
        schema_url: Optional[str] = None,
    ) -> None:
        super().__init__(name, version=version, schema_url=schema_url)
        self._lock = Lock()
        self._instruments: List[_ProxyInstrumentT] = []
        self._real_meter: Optional[Meter] = None

    def on_set_meter_provider(self, meter_provider: MeterProvider) -> None:
        """Called when a real meter provider is set on the creating _ProxyMeterProvider

        Creates a real backing meter for this instance and notifies all created
        instruments so they can create real backing instruments.
        """
        real_meter = meter_provider.get_meter(
            self._name, self._version, self._schema_url
        )

        with self._lock:
            self._real_meter = real_meter
            # notify all proxy instruments of the new meter so they can create
            # real instruments to back themselves
            for instrument in self._instruments:
                instrument.on_meter_set(real_meter)

    def create_counter(
        self,
        name: str,
        unit: str = "",
        description: str = "",
    ) -> Counter:
        with self._lock:
            if self._real_meter:
                return self._real_meter.create_counter(name, unit, description)
            proxy = _ProxyCounter(name, unit, description)
            self._instruments.append(proxy)
            return proxy

    def create_up_down_counter(
        self,
        name: str,
        unit: str = "",
        description: str = "",
    ) -> UpDownCounter:
        with self._lock:
            if self._real_meter:
                return self._real_meter.create_up_down_counter(
                    name, unit, description
                )
            proxy = _ProxyUpDownCounter(name, unit, description)
            self._instruments.append(proxy)
            return proxy

    def create_observable_counter(
        self,
        name: str,
        callbacks: Optional[Sequence[CallbackT]] = None,
        unit: str = "",
        description: str = "",
    ) -> ObservableCounter:
        with self._lock:
            if self._real_meter:
                return self._real_meter.create_observable_counter(
                    name, callbacks, unit, description
                )
            proxy = _ProxyObservableCounter(
                name, callbacks, unit=unit, description=description
            )
            self._instruments.append(proxy)
            return proxy

    def create_histogram(
        self,
        name: str,
        unit: str = "",
        description: str = "",
    ) -> Histogram:
        with self._lock:
            if self._real_meter:
                return self._real_meter.create_histogram(
                    name, unit, description
                )
            proxy = _ProxyHistogram(name, unit, description)
            self._instruments.append(proxy)
            return proxy

    def create_observable_gauge(
        self,
        name: str,
        callbacks: Optional[Sequence[CallbackT]] = None,
        unit: str = "",
        description: str = "",
    ) -> ObservableGauge:
        with self._lock:
            if self._real_meter:
                return self._real_meter.create_observable_gauge(
                    name, callbacks, unit, description
                )
            proxy = _ProxyObservableGauge(
                name, callbacks, unit=unit, description=description
            )
            self._instruments.append(proxy)
            return proxy

    def create_observable_up_down_counter(
        self,
        name: str,
        callbacks: Optional[Sequence[CallbackT]] = None,
        unit: str = "",
        description: str = "",
    ) -> ObservableUpDownCounter:
        with self._lock:
            if self._real_meter:
                return self._real_meter.create_observable_up_down_counter(
                    name,
                    callbacks,
                    unit,
                    description,
                )
            proxy = _ProxyObservableUpDownCounter(
                name, callbacks, unit=unit, description=description
            )
            self._instruments.append(proxy)
            return proxy


class NoOpMeter(Meter):
    """The default Meter used when no Meter implementation is available.

    All operations are no-op.
    """

    def create_counter(
        self,
        name: str,
        unit: str = "",
        description: str = "",
    ) -> Counter:
        """Returns a no-op Counter."""
        super().create_counter(name, unit=unit, description=description)
        if self._is_instrument_registered(
            name, NoOpCounter, unit, description
        )[0]:
            _logger.warning(
                "An instrument with name %s, type %s, unit %s and "
                "description %s has been created already.",
                name,
                Counter.__name__,
                unit,
                description,
            )
        return NoOpCounter(name, unit=unit, description=description)

    def create_up_down_counter(
        self,
        name: str,
        unit: str = "",
        description: str = "",
    ) -> UpDownCounter:
        """Returns a no-op UpDownCounter."""
        super().create_up_down_counter(
            name, unit=unit, description=description
        )
        if self._is_instrument_registered(
            name, NoOpUpDownCounter, unit, description
        )[0]:
            _logger.warning(
                "An instrument with name %s, type %s, unit %s and "
                "description %s has been created already.",
                name,
                UpDownCounter.__name__,
                unit,
                description,
            )
        return NoOpUpDownCounter(name, unit=unit, description=description)

    def create_observable_counter(
        self,
        name: str,
        callbacks: Optional[Sequence[CallbackT]] = None,
        unit: str = "",
        description: str = "",
    ) -> ObservableCounter:
        """Returns a no-op ObservableCounter."""
        super().create_observable_counter(
            name, callbacks, unit=unit, description=description
        )
        if self._is_instrument_registered(
            name, NoOpObservableCounter, unit, description
        )[0]:
            _logger.warning(
                "An instrument with name %s, type %s, unit %s and "
                "description %s has been created already.",
                name,
                ObservableCounter.__name__,
                unit,
                description,
            )
        return NoOpObservableCounter(
            name,
            callbacks,
            unit=unit,
            description=description,
        )

    def create_histogram(
        self,
        name: str,
        unit: str = "",
        description: str = "",
    ) -> Histogram:
        """Returns a no-op Histogram."""
        super().create_histogram(name, unit=unit, description=description)
        if self._is_instrument_registered(
            name, NoOpHistogram, unit, description
        )[0]:
            _logger.warning(
                "An instrument with name %s, type %s, unit %s and "
                "description %s has been created already.",
                name,
                Histogram.__name__,
                unit,
                description,
            )
        return NoOpHistogram(name, unit=unit, description=description)

    def create_observable_gauge(
        self,
        name: str,
        callbacks: Optional[Sequence[CallbackT]] = None,
        unit: str = "",
        description: str = "",
    ) -> ObservableGauge:
        """Returns a no-op ObservableGauge."""
        super().create_observable_gauge(
            name, callbacks, unit=unit, description=description
        )
        if self._is_instrument_registered(
            name, NoOpObservableGauge, unit, description
        )[0]:
            _logger.warning(
                "An instrument with name %s, type %s, unit %s and "
                "description %s has been created already.",
                name,
                ObservableGauge.__name__,
                unit,
                description,
            )
        return NoOpObservableGauge(
            name,
            callbacks,
            unit=unit,
            description=description,
        )

    def create_observable_up_down_counter(
        self,
        name: str,
        callbacks: Optional[Sequence[CallbackT]] = None,
        unit: str = "",
        description: str = "",
    ) -> ObservableUpDownCounter:
        """Returns a no-op ObservableUpDownCounter."""
        super().create_observable_up_down_counter(
            name, callbacks, unit=unit, description=description
        )
        if self._is_instrument_registered(
            name, NoOpObservableUpDownCounter, unit, description
        )[0]:
            _logger.warning(
                "An instrument with name %s, type %s, unit %s and "
                "description %s has been created already.",
                name,
                ObservableUpDownCounter.__name__,
                unit,
                description,
            )
        return NoOpObservableUpDownCounter(
            name,
            callbacks,
            unit=unit,
            description=description,
        )


_METER_PROVIDER_SET_ONCE = Once()
_METER_PROVIDER: Optional[MeterProvider] = None
_PROXY_METER_PROVIDER = _ProxyMeterProvider()


def get_meter(
    name: str,
    version: str = "",
    meter_provider: Optional[MeterProvider] = None,
) -> "Meter":
    """Returns a `Meter` for use by the given instrumentation library.

    This function is a convenience wrapper for
    `opentelemetry.metrics.MeterProvider.get_meter`.

    If meter_provider is omitted the current configured one is used.
    """
    if meter_provider is None:
        meter_provider = get_meter_provider()
    return meter_provider.get_meter(name, version)


def _set_meter_provider(meter_provider: MeterProvider, log: bool) -> None:
    def set_mp() -> None:
        global _METER_PROVIDER  # pylint: disable=global-statement
        _METER_PROVIDER = meter_provider

        # gives all proxies real instruments off the newly set meter provider
        _PROXY_METER_PROVIDER.on_set_meter_provider(meter_provider)

    did_set = _METER_PROVIDER_SET_ONCE.do_once(set_mp)

    if log and not did_set:
        _logger.warning("Overriding of current MeterProvider is not allowed")


def set_meter_provider(meter_provider: MeterProvider) -> None:
    """Sets the current global :class:`~.MeterProvider` object.

    This can only be done once, a warning will be logged if any further attempt
    is made.
    """
    _set_meter_provider(meter_provider, log=True)


def get_meter_provider() -> MeterProvider:
    """Gets the current global :class:`~.MeterProvider` object."""

    if _METER_PROVIDER is None:
        if OTEL_PYTHON_METER_PROVIDER not in environ.keys():
            return _PROXY_METER_PROVIDER

        meter_provider: MeterProvider = _load_provider(  # type: ignore
            OTEL_PYTHON_METER_PROVIDER, "meter_provider"
        )
        _set_meter_provider(meter_provider, log=False)

    # _METER_PROVIDER will have been set by one thread
    return cast("MeterProvider", _METER_PROVIDER)



class EventManagerException(Exception):
    pass


class CannotProcessSyncEvent(EventManagerException):
    pass


class AsyncTargetForEventWithNoData(EventManagerException):
    pass


class AsyncTargetForEventWithNoUtf8Data(EventManagerException):
    pass


class AsyncTargetForEventWithNoJsonData(EventManagerException):
    pass


class AsyncTargetForEventWithDataWithoutId(EventManagerException):
    pass


class CannotProcessATargetAndAsyncTarget(EventManagerException):
    pass



"""distutils.pypirc

Provides the PyPIRCCommand class, the base class for the command classes
that uses .pypirc in the distutils.command package.
"""

import email.message
import os
from configparser import RawConfigParser

from .cmd import Command

DEFAULT_PYPIRC = """\
[distutils]
index-servers =
    pypi

[pypi]
username:%s
password:%s
"""


class PyPIRCCommand(Command):
    """Base command that knows how to handle the .pypirc file"""

    DEFAULT_REPOSITORY = 'https://upload.pypi.org/legacy/'
    DEFAULT_REALM = 'pypi'
    repository = None
    realm = None

    user_options = [
        ('repository=', 'r', "url of repository [default: %s]" % DEFAULT_REPOSITORY),
        ('show-response', None, 'display full response text from server'),
    ]

    boolean_options = ['show-response']

    def _get_rc_file(self):
        """Returns rc file path."""
        return os.path.join(os.path.expanduser('~'), '.pypirc')

    def _store_pypirc(self, username, password):
        """Creates a default .pypirc file."""
        rc = self._get_rc_file()
        raw = os.open(rc, os.O_CREAT | os.O_WRONLY, 0o600)
        with os.fdopen(raw, 'w', encoding='utf-8') as f:
            f.write(DEFAULT_PYPIRC % (username, password))

    def _read_pypirc(self):  # noqa: C901
        """Reads the .pypirc file."""
        rc = self._get_rc_file()
        if os.path.exists(rc):
            self.announce('Using PyPI login from %s' % rc)
            repository = self.repository or self.DEFAULT_REPOSITORY

            config = RawConfigParser()
            config.read(rc, encoding='utf-8')
            sections = config.sections()
            if 'distutils' in sections:
                # let's get the list of servers
                index_servers = config.get('distutils', 'index-servers')
                _servers = [
                    server.strip()
                    for server in index_servers.split('\n')
                    if server.strip() != ''
                ]
                if _servers == []:
                    # nothing set, let's try to get the default pypi
                    if 'pypi' in sections:
                        _servers = ['pypi']
                    else:
                        # the file is not properly defined, returning
                        # an empty dict
                        return {}
                for server in _servers:
                    current = {'server': server}
                    current['username'] = config.get(server, 'username')

                    # optional params
                    for key, default in (
                        ('repository', self.DEFAULT_REPOSITORY),
                        ('realm', self.DEFAULT_REALM),
                        ('password', None),
                    ):
                        if config.has_option(server, key):
                            current[key] = config.get(server, key)
                        else:
                            current[key] = default

                    # work around people having "repository" for the "pypi"
                    # section of their config set to the HTTP (rather than
                    # HTTPS) URL
                    if server == 'pypi' and repository in (
                        self.DEFAULT_REPOSITORY,
                        'pypi',
                    ):
                        current['repository'] = self.DEFAULT_REPOSITORY
                        return current

                    if (
                        current['server'] == repository
                        or current['repository'] == repository
                    ):
                        return current
            elif 'server-login' in sections:
                # old format
                server = 'server-login'
                if config.has_option(server, 'repository'):
                    repository = config.get(server, 'repository')
                else:
                    repository = self.DEFAULT_REPOSITORY
                return {
                    'username': config.get(server, 'username'),
                    'password': config.get(server, 'password'),
                    'repository': repository,
                    'server': server,
                    'realm': self.DEFAULT_REALM,
                }

        return {}

    def _read_pypi_response(self, response):
        """Read and decode a PyPI HTTP response."""
        content_type = response.getheader('content-type', 'text/plain')
        return response.read().decode(_extract_encoding(content_type))

    def initialize_options(self):
        """Initialize options."""
        self.repository = None
        self.realm = None
        self.show_response = 0

    def finalize_options(self):
        """Finalizes options."""
        if self.repository is None:
            self.repository = self.DEFAULT_REPOSITORY
        if self.realm is None:
            self.realm = self.DEFAULT_REALM


def _extract_encoding(content_type):
    """
    >>> _extract_encoding('text/plain')
    'ascii'
    >>> _extract_encoding('text/html; charset="utf8"')
    'utf8'
    """
    msg = email.message.EmailMessage()
    msg['content-type'] = content_type
    return msg['content-type'].params.get('charset', 'ascii')



# ik.py - functions for handling Estonian Personal ID numbers (IK)
# coding: utf-8
#
# Copyright (C) 2015 Tomas Karasek
# Copyright (C) 2015 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Isikukood (Estonian Personal ID number).

The number consists of 11 digits: the first indicates the gender and century
the person was born in, the following 6 digits the birth date, followed by a
3 digit serial and a check digit.

More information:

* https://www.riigiteataja.ee/akt/106032012004

>>> validate('36805280109')
'36805280109'
>>> validate('36805280108')  # incorrect check digit
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> get_birth_date('36805280109')
datetime.date(1968, 5, 28)
"""

import datetime

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' ').strip()


def get_birth_date(number):
    """Split the date parts from the number and return the birth date."""
    number = compact(number)
    if number[0] in '12':
        century = 1800
    elif number[0] in '34':
        century = 1900
    elif number[0] in '56':
        century = 2000
    elif number[0] in '78':
        century = 2100
    else:
        raise InvalidComponent()
    year = century + int(number[1:3])
    month = int(number[3:5])
    day = int(number[5:7])
    try:
        return datetime.date(year, month, day)
    except ValueError:
        raise InvalidComponent()


def get_gender(number):
    """Get the person's birth gender ('M' or 'F')."""
    number = compact(number)
    if number[0] in '1357':
        return 'M'
    elif number[0] in '2468':
        return 'F'
    else:
        raise InvalidComponent()


def calc_check_digit(number):
    """Calculate the check digit."""
    check = sum(((i % 9) + 1) * int(n)
                for i, n in enumerate(number[:-1])) % 11
    if check == 10:
        check = sum((((i + 2) % 9) + 1) * int(n)
                    for i, n in enumerate(number[:-1])) % 11
    return str(check % 10)


def validate(number):
    """Check if the number provided is valid. This checks the length,
    formatting, embedded date and check digit."""
    number = compact(number)
    if not isdigits(number):
        raise InvalidFormat()
    if len(number) != 11:
        raise InvalidLength()
    get_birth_date(number)
    if number[-1] != calc_check_digit(number):
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number provided is valid. This checks the length,
    formatting, embedded date and check digit."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



from production.box import DistributorBox
from production.treebuilder import BoxConfiguration

from ._BoxBuilder import BoxBuilder


class DistributorBoxBuilder(BoxBuilder):
    def __init__(self, loggers):
        super(DistributorBoxBuilder, self).__init__(loggers)

    def build(self, config):
        # type: (BoxConfiguration) -> DistributorBox
        return DistributorBox(config.name, config.extra_config["Paths"], self.get_logger(config))



# -*- coding: utf-8 -*-

import logging

from typing import (
    Any,
    Dict,
)

from old_helper import (
    remove_punctuation,
    validar_cnpj,
)
from helper import remove_accents
from pos_model import Order, TenderType, OrderTender

from models import PosTenderToNfeTender, NfeTender
from _NfceXmlPartBuilder import NfceXmlPartBuilder

logger = logging.getLogger("FiscalWrapper")
CARD_XML = "<card><tpIntegra>{0}</tpIntegra><CNPJ>{1}</CNPJ><tBand>{2}</tBand>{3}</card>"
CARD_XML_INVALID_CNPJ = "<card><tpIntegra>2</tpIntegra></card>"


class PagBuilder(NfceXmlPartBuilder):

    def __init__(self, versao_ws, tenders_type):
        # type: (int, Dict) -> None

        self.versao_ws = versao_ws
        self.tenders_type = tenders_type

    def build_xml(self, order, context):
        # type: (Order, Dict[unicode, Any]) -> unicode

        xml = "<pag>" if self.versao_ws not in (1, 3) else ""

        for tender in order.tenders:
            xml += "<detPag>" if self.versao_ws not in (1, 3) else "<pag>"

            tender_type = self._get_tender_type(tender)
            xml += "<tPag>{:>02}</tPag>".format(tender_type)

            if tender_type == NfeTender.outros:
                tender_desc = tender.descr or tender.type
                xml += "<xPag>{}</xPag>".format(remove_accents(tender_desc))

            total_amount = tender.amount - tender.change if self.versao_ws in (1, 3) else tender.amount
            xml += "<vPag>{:>0.2f}</vPag>".format(total_amount)

            xml += self._get_card_xml(tender)
            xml += "</detPag>" if self.versao_ws not in (1, 3) else "</pag>"

            if self.versao_ws not in (1, 3) and tender.change > 0.0:
                xml += "<vTroco>{:>0.2f}</vTroco>".format(tender.change)

        xml += "</pag>" if self.versao_ws not in (1, 3) else ""

        return xml

    def _get_card_xml(
        self,
        tender  # type: OrderTender
    ):
        # type: (...) -> str

        if self._is_tef_tender(tender):
            if not validar_cnpj(tender.cnpj_auth):
                return CARD_XML_INVALID_CNPJ

            card_brand = self._get_card_brand(tender)

            cnpj_auth = remove_punctuation(tender.cnpj_auth.strip())
            auth_code = "<cAut>{0}</cAut>".format(tender.auth_code.strip()[-20:])
            integrated = 1
            return CARD_XML.format(integrated, cnpj_auth, card_brand, auth_code)

        tender_description = str(remove_accents(tender.descr)).upper()
        if tender_description in self.tenders_type:
            tender_type_data_base = self.tenders_type[tender_description]
            if self._tender_database_is_valid(tender_type_data_base):
                card_brand = tender_type_data_base.get("tBand", None)
                card_brand = card_brand if self._is_not_blank(card_brand) else "99"
                cnpj_auth = remove_punctuation(tender_type_data_base.get("cnpj", None))
                not_integrated = 2
                return CARD_XML.format(not_integrated, cnpj_auth, card_brand, "")

        return ""

    @staticmethod
    def _get_card_brand(
        tender,     # type: OrderTender
    ):
        card_brand = tender.t_band or "99"
        try:
            int(card_brand)
        except ValueError:
            logger.warning("Invalid Card Brand {}".format(str(card_brand)))
            card_brand = "99"

        return card_brand

    def _get_tender_type(
        self,
        tender  # type: OrderTender
    ):
        # type: (...) -> int

        if self._is_tef_tender(tender):
            return int(tender.t_pag or 99)

        tender_description = str(remove_accents(tender.descr)).upper()
        if tender_description in self.tenders_type:
            tender_type_data_base = self.tenders_type[tender_description]
            t_pag = tender_type_data_base.get("tPag", None)
            if self._tender_database_is_valid(tender_type_data_base) and self._is_not_blank(t_pag):
                return int(t_pag)

        tender_type = PosTenderToNfeTender.convert(tender.type)
        if tender_type not in [NfeTender.dinheiro, NfeTender.cash_back]:
            return NfeTender.outros

        return tender_type

    @staticmethod
    def _is_tef_tender(
        tender  # type: OrderTender
    ):
        # type: (...) -> bool

        tef_tenders = [
            TenderType.credit,
            TenderType.debit,
            TenderType.voucher,
            TenderType.gift_card,
            TenderType.digital_wallet,
        ]
        return tender.type in tef_tenders

    @staticmethod
    def _tender_database_is_valid(
        tender_type_data_base   # type: Dict
    ):
        # type: (...) -> bool

        card_cnpj = tender_type_data_base.get("cnpj", None)
        cnpj = validar_cnpj(card_cnpj)
        return True if cnpj is not False else False

    @staticmethod
    def _is_not_blank(
        value   # type: str
    ):
        # type: (...) -> bool

        if value and value.strip():
            return True
        return False



# Copyright (C) Dnspython Contributors, see LICENSE for text of ISC license

# Copyright (C) 2006-2017 Nominum, Inc.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose with or without fee is hereby granted,
# provided that the above copyright notice and this permission notice
# appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND NOMINUM DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL NOMINUM BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""DNS E.164 helpers."""

import dns.exception
import dns.name
import dns.resolver
from ._compat import string_types, maybe_decode

#: The public E.164 domain.
public_enum_domain = dns.name.from_text('e164.arpa.')


def from_e164(text, origin=public_enum_domain):
    """Convert an E.164 number in textual form into a Name object whose
    value is the ENUM domain name for that number.

    Non-digits in the text are ignored, i.e. "16505551212",
    "+1.650.555.1212" and "1 (650) 555-1212" are all the same.

    *text*, a ``text``, is an E.164 number in textual form.

    *origin*, a ``dns.name.Name``, the domain in which the number
    should be constructed.  The default is ``e164.arpa.``.

    Returns a ``dns.name.Name``.
    """

    parts = [d for d in text if d.isdigit()]
    parts.reverse()
    return dns.name.from_text('.'.join(parts), origin=origin)


def to_e164(name, origin=public_enum_domain, want_plus_prefix=True):
    """Convert an ENUM domain name into an E.164 number.

    Note that dnspython does not have any information about preferred
    number formats within national numbering plans, so all numbers are
    emitted as a simple string of digits, prefixed by a '+' (unless
    *want_plus_prefix* is ``False``).

    *name* is a ``dns.name.Name``, the ENUM domain name.

    *origin* is a ``dns.name.Name``, a domain containing the ENUM
    domain name.  The name is relativized to this domain before being
    converted to text.  If ``None``, no relativization is done.

    *want_plus_prefix* is a ``bool``.  If True, add a '+' to the beginning of
    the returned number.

    Returns a ``text``.

    """
    if origin is not None:
        name = name.relativize(origin)
    dlabels = [d for d in name.labels if d.isdigit() and len(d) == 1]
    if len(dlabels) != len(name.labels):
        raise dns.exception.SyntaxError('non-digit labels in ENUM domain name')
    dlabels.reverse()
    text = b''.join(dlabels)
    if want_plus_prefix:
        text = b'+' + text
    return maybe_decode(text)


def query(number, domains, resolver=None):
    """Look for NAPTR RRs for the specified number in the specified domains.

    e.g. lookup('16505551212', ['e164.dnspython.org.', 'e164.arpa.'])

    *number*, a ``text`` is the number to look for.

    *domains* is an iterable containing ``dns.name.Name`` values.

    *resolver*, a ``dns.resolver.Resolver``, is the resolver to use.  If
    ``None``, the default resolver is used.
    """

    if resolver is None:
        resolver = dns.resolver.get_default_resolver()
    e_nx = dns.resolver.NXDOMAIN()
    for domain in domains:
        if isinstance(domain, string_types):
            domain = dns.name.from_text(domain)
        qname = dns.e164.from_e164(number, domain)
        try:
            return resolver.query(qname, 'NAPTR')
        except dns.resolver.NXDOMAIN as e:
            e_nx += e
    raise e_nx



from pymaging import Image
from pymaging.colors import RGB
from pymaging.formats import registry
from pymaging.shapes import Line
from pymaging.webcolors import Black, White
from pymaging_png.png import PNG

import qrcode.image.base


class PymagingImage(qrcode.image.base.BaseImage):
    """
    pymaging image builder, default format is PNG.
    """
    kind = "PNG"
    allowed_kinds = ("PNG",)

    def __init__(self, *args, **kwargs):
        """
        Register PNG with pymaging.
        """
        registry.formats = []
        registry.names = {}
        registry._populate()
        registry.register(PNG)

        super(PymagingImage, self).__init__(*args, **kwargs)

    def new_image(self, **kwargs):
        return Image.new(RGB, self.pixel_size, self.pixel_size, White)

    def drawrect(self, row, col):
        (x, y), (x2, y2) = self.pixel_box(row, col)
        for r in range(self.box_size):
            line_y = y + r
            line = Line(x, line_y, x2, line_y)
            self._img.draw(line, Black)

    def save(self, stream, kind=None):
        self._img.save(stream, self.check_kind(kind))

    def check_kind(self, kind, transform=None, **kwargs):
        """
        pymaging (pymaging_png at least) uses lower case for the type.
        """
        if transform is None:
            transform = lambda x: x.lower()
        return super(PymagingImage, self).check_kind(
            kind, transform=transform, **kwargs)



# Copyright The OpenTelemetry Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gzip
import logging
import zlib
from os import environ
from typing import Dict, Optional, Any, Callable, List
from typing import Sequence, Mapping  # noqa: F401

from io import BytesIO
from time import sleep
from deprecated import deprecated

from opentelemetry.exporter.otlp.proto.common._internal import (
    _get_resource_data,
    _create_exp_backoff_generator,
)
from opentelemetry.exporter.otlp.proto.common._internal.metrics_encoder import (
    OTLPMetricExporterMixin,
)
from opentelemetry.exporter.otlp.proto.common.metrics_encoder import (
    encode_metrics,
)
from opentelemetry.exporter.otlp.proto.http import Compression
from opentelemetry.sdk.metrics._internal.aggregation import Aggregation
from opentelemetry.proto.collector.metrics.v1.metrics_service_pb2 import (  # noqa: F401
    ExportMetricsServiceRequest,
)
from opentelemetry.proto.common.v1.common_pb2 import (  # noqa: F401
    AnyValue,
    ArrayValue,
    KeyValue,
    KeyValueList,
)
from opentelemetry.proto.common.v1.common_pb2 import (  # noqa: F401
    InstrumentationScope,
)
from opentelemetry.proto.resource.v1.resource_pb2 import Resource  # noqa: F401
from opentelemetry.proto.metrics.v1 import metrics_pb2 as pb2  # noqa: F401
from opentelemetry.sdk.environment_variables import (
    OTEL_EXPORTER_OTLP_ENDPOINT,
    OTEL_EXPORTER_OTLP_CERTIFICATE,
    OTEL_EXPORTER_OTLP_HEADERS,
    OTEL_EXPORTER_OTLP_TIMEOUT,
    OTEL_EXPORTER_OTLP_COMPRESSION,
    OTEL_EXPORTER_OTLP_METRICS_ENDPOINT,
    OTEL_EXPORTER_OTLP_METRICS_CERTIFICATE,
    OTEL_EXPORTER_OTLP_METRICS_HEADERS,
    OTEL_EXPORTER_OTLP_METRICS_TIMEOUT,
    OTEL_EXPORTER_OTLP_METRICS_COMPRESSION,
)
from opentelemetry.sdk.metrics.export import (
    AggregationTemporality,
    MetricExporter,
    MetricExportResult,
    MetricsData,
)
from opentelemetry.sdk.metrics.export import (  # noqa: F401
    Gauge,
    Histogram as HistogramType,
    Sum,
)
from opentelemetry.sdk.resources import Resource as SDKResource
from opentelemetry.util.re import parse_env_headers

import requests
from opentelemetry.proto.resource.v1.resource_pb2 import (
    Resource as PB2Resource,
)

_logger = logging.getLogger(__name__)


DEFAULT_COMPRESSION = Compression.NoCompression
DEFAULT_ENDPOINT = "http://localhost:4318/"
DEFAULT_METRICS_EXPORT_PATH = "v1/metrics"
DEFAULT_TIMEOUT = 10  # in seconds


class OTLPMetricExporter(MetricExporter, OTLPMetricExporterMixin):

    _MAX_RETRY_TIMEOUT = 64

    def __init__(
        self,
        endpoint: Optional[str] = None,
        certificate_file: Optional[str] = None,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[int] = None,
        compression: Optional[Compression] = None,
        session: Optional[requests.Session] = None,
        preferred_temporality: Dict[type, AggregationTemporality] = None,
        preferred_aggregation: Dict[type, Aggregation] = None,
    ):
        self._endpoint = endpoint or environ.get(
            OTEL_EXPORTER_OTLP_METRICS_ENDPOINT,
            _append_metrics_path(
                environ.get(OTEL_EXPORTER_OTLP_ENDPOINT, DEFAULT_ENDPOINT)
            ),
        )
        self._certificate_file = certificate_file or environ.get(
            OTEL_EXPORTER_OTLP_METRICS_CERTIFICATE,
            environ.get(OTEL_EXPORTER_OTLP_CERTIFICATE, True),
        )
        headers_string = environ.get(
            OTEL_EXPORTER_OTLP_METRICS_HEADERS,
            environ.get(OTEL_EXPORTER_OTLP_HEADERS, ""),
        )
        self._headers = headers or parse_env_headers(headers_string)
        self._timeout = timeout or int(
            environ.get(
                OTEL_EXPORTER_OTLP_METRICS_TIMEOUT,
                environ.get(OTEL_EXPORTER_OTLP_TIMEOUT, DEFAULT_TIMEOUT),
            )
        )
        self._compression = compression or _compression_from_env()
        self._session = session or requests.Session()
        self._session.headers.update(self._headers)
        self._session.headers.update(
            {"Content-Type": "application/x-protobuf"}
        )
        if self._compression is not Compression.NoCompression:
            self._session.headers.update(
                {"Content-Encoding": self._compression.value}
            )

        self._common_configuration(preferred_temporality)

    def _export(self, serialized_data: str):
        data = serialized_data
        if self._compression == Compression.Gzip:
            gzip_data = BytesIO()
            with gzip.GzipFile(fileobj=gzip_data, mode="w") as gzip_stream:
                gzip_stream.write(serialized_data)
            data = gzip_data.getvalue()
        elif self._compression == Compression.Deflate:
            data = zlib.compress(bytes(serialized_data))

        return self._session.post(
            url=self._endpoint,
            data=data,
            verify=self._certificate_file,
            timeout=self._timeout,
        )

    @staticmethod
    def _retryable(resp: requests.Response) -> bool:
        if resp.status_code == 408:
            return True
        if resp.status_code >= 500 and resp.status_code <= 599:
            return True
        return False

    def export(
        self,
        metrics_data: MetricsData,
        timeout_millis: float = 10_000,
        **kwargs,
    ) -> MetricExportResult:
        serialized_data = encode_metrics(metrics_data)
        for delay in _create_exp_backoff_generator(
            max_value=self._MAX_RETRY_TIMEOUT
        ):

            if delay == self._MAX_RETRY_TIMEOUT:
                return MetricExportResult.FAILURE

            resp = self._export(serialized_data.SerializeToString())
            # pylint: disable=no-else-return
            if resp.status_code in (200, 202):
                return MetricExportResult.SUCCESS
            elif self._retryable(resp):
                _logger.warning(
                    "Transient error %s encountered while exporting metric batch, retrying in %ss.",
                    resp.reason,
                    delay,
                )
                sleep(delay)
                continue
            else:
                _logger.error(
                    "Failed to export batch code: %s, reason: %s",
                    resp.status_code,
                    resp.text,
                )
                return MetricExportResult.FAILURE
        return MetricExportResult.FAILURE

    def shutdown(self, timeout_millis: float = 30_000, **kwargs) -> None:
        pass

    @property
    def _exporting(self) -> str:
        return "metrics"

    def force_flush(self, timeout_millis: float = 10_000) -> bool:
        """Nothing is buffered in this exporter, so this method does nothing."""
        return True


@deprecated(
    version="1.18.0",
    reason="Use one of the encoders from opentelemetry-exporter-otlp-proto-common instead",
)
def get_resource_data(
    sdk_resource_scope_data: Dict[SDKResource, Any],  # ResourceDataT?
    resource_class: Callable[..., PB2Resource],
    name: str,
) -> List[PB2Resource]:
    return _get_resource_data(sdk_resource_scope_data, resource_class, name)


def _compression_from_env() -> Compression:
    compression = (
        environ.get(
            OTEL_EXPORTER_OTLP_METRICS_COMPRESSION,
            environ.get(OTEL_EXPORTER_OTLP_COMPRESSION, "none"),
        )
        .lower()
        .strip()
    )
    return Compression(compression)


def _append_metrics_path(endpoint: str) -> str:
    if endpoint.endswith("/"):
        return endpoint + DEFAULT_METRICS_EXPORT_PATH
    return endpoint + f"/{DEFAULT_METRICS_EXPORT_PATH}"



from messagebus import TokenCreator, TokenPriority


MSGGRP_SLCTRL_ORDERMGR = "A"

TK_SLCTRL_OMGR_DOSALE = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=MSGGRP_SLCTRL_ORDERMGR,
    number="1",
)

TK_SLCTRL_OMGR_ORDERPICT = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=MSGGRP_SLCTRL_ORDERMGR,
    number="2"
)

TK_SLCTRL_OMGR_CREATEORDER = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=MSGGRP_SLCTRL_ORDERMGR,
    number="18"
)

TK_SLCTRL_OMGR_LISTORDERS = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=MSGGRP_SLCTRL_ORDERMGR,
    number="D"
)

TK_SLCTRL_OMGR_VOIDLINE = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=MSGGRP_SLCTRL_ORDERMGR,
    number="3",
)

TK_SLCTRL_OMGR_DOOPTION = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=MSGGRP_SLCTRL_ORDERMGR,
    number="A",
)

TK_SLCTRL_OMGR_CHANGEMODIFIER = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=MSGGRP_SLCTRL_ORDERMGR,
    number="13",
)


def get_tokens():
    return (
        TK_SLCTRL_OMGR_CHANGEMODIFIER,
        TK_SLCTRL_OMGR_DOOPTION,
        TK_SLCTRL_OMGR_DOSALE,
        TK_SLCTRL_OMGR_ORDERPICT,
        TK_SLCTRL_OMGR_CREATEORDER,
        TK_SLCTRL_OMGR_LISTORDERS,
        TK_SLCTRL_OMGR_VOIDLINE,
    )



from functions_authorization import get_function_authorization_level_from_config
from get_keyboard_comments_config import get_keyboard_comments_config
from get_min_value_to_ask_bordereau_justify_config import get_min_value_to_ask_bordereau_justify_config
from show_numpad_for_customer_name import show_numpad_for_customer_name
from get_production_config import get_hold_time_config
from should_ask_pager_number import should_ask_pager_number
from should_ask_delivery_type import should_ask_delivery_type
from get_default_delivery_type import get_default_delivery_type
from get_pre_account_allow_printing import get_pre_account_allow_printing
from get_hide_unpriced_items_from_total_report import get_hide_unpriced_items_from_total_report
from get_pre_account_with_discount import get_pre_account_with_discount
from can_cancel_tef_payment_on_error import can_cancel_tef_payment_on_error
from ._DeliveryCoupon import must_print_delivery_coupon
from should_ask_loyalty_message import (
    should_ask_loyalty_message_on_total,
    should_ask_loyalty_on_payment,
)
from can_create_table_parallel_order import can_create_table_parallel_order
from get_close_operator_config import get_close_operator_conference_type
from considers_tef_on_bordereau import considers_tef_on_bordereau



from threading import Condition

from ._AwakableSleeper import AwakableSleeper


class ConditionSleeper(AwakableSleeper):
    def __init__(self):
        self.condition = Condition()

    def sleep(self, secs):
        with self.condition:
            self.condition.wait(secs)

    def wake_up(self):
        with self.condition:
            self.condition.notify_all()



#! python
#
# This module implements a special URL handler that allows selecting an
# alternate implementation provided by some backends.
#
# This file is part of pySerial. https://github.com/pyserial/pyserial
# (C) 2015 Chris Liechti <cliechti@gmx.net>
#
# SPDX-License-Identifier:    BSD-3-Clause
#
# URL format:    alt://port[?option[=value][&option[=value]]]
# options:
# - class=X used class named X instead of Serial
#
# example:
#   use poll based implementation on Posix (Linux):
#   python -m serial.tools.miniterm alt:///dev/ttyUSB0?class=PosixPollSerial

from __future__ import absolute_import

try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse

import serial


def serial_class_for_url(url):
    """extract host and port from an URL string"""
    parts = urlparse.urlsplit(url)
    if parts.scheme != 'alt':
        raise serial.SerialException(
            'expected a string in the form "alt://port[?option[=value][&option[=value]]]": '
            'not starting with alt:// ({!r})'.format(parts.scheme))
    class_name = 'Serial'
    try:
        for option, values in urlparse.parse_qs(parts.query, True).items():
            if option == 'class':
                class_name = values[0]
            else:
                raise ValueError('unknown option: {!r}'.format(option))
    except ValueError as e:
        raise serial.SerialException(
            'expected a string in the form '
            '"alt://port[?option[=value][&option[=value]]]": {!r}'.format(e))
    if not hasattr(serial, class_name):
        raise ValueError('unknown class: {!r}'.format(class_name))
    cls = getattr(serial, class_name)
    if not issubclass(cls, serial.Serial):
        raise ValueError('class {!r} is not an instance of Serial'.format(class_name))
    return (''.join([parts.netloc, parts.path]), cls)

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
if __name__ == '__main__':
    s = serial.serial_for_url('alt:///dev/ttyS0?class=PosixPollSerial')
    print(s)



import os

import cfgtools

hide_unpriced_items_from_total_report = None


def get_hide_unpriced_items_from_total_report():
    # type: (...) -> bool

    global hide_unpriced_items_from_total_report

    if hide_unpriced_items_from_total_report is None:
        config = cfgtools.read(os.environ["LOADERCFG"])
        hide_unpriced_items_from_total_report = config.find_value("Customizations.HideUnpricedItemsFromTotalReport") or "true"

    return hide_unpriced_items_from_total_report.lower() == "true"



class ProductServiceException(Exception):
    pass


class InvalidPartCode(Exception):
    def __init__(
        self,           # type: InvalidPartCode
        part_code,      # type: int
    ):
        # type: (...) -> None

        self.part_code = part_code



# Copyright (C) Dnspython Contributors, see LICENSE for text of ISC license

# Copyright (C) 2004-2007, 2009-2011 Nominum, Inc.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose with or without fee is hereby granted,
# provided that the above copyright notice and this permission notice
# appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND NOMINUM DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL NOMINUM BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import struct

import dns.exception
import dns.rdata
import dns.rdatatype
import dns.name
from dns._compat import xrange


class NSEC(dns.rdata.Rdata):

    """NSEC record

    @ivar next: the next name
    @type next: dns.name.Name object
    @ivar windows: the windowed bitmap list
    @type windows: list of (window number, string) tuples"""

    __slots__ = ['next', 'windows']

    def __init__(self, rdclass, rdtype, next, windows):
        super(NSEC, self).__init__(rdclass, rdtype)
        self.next = next
        self.windows = windows

    def to_text(self, origin=None, relativize=True, **kw):
        next = self.next.choose_relativity(origin, relativize)
        text = ''
        for (window, bitmap) in self.windows:
            bits = []
            for i in xrange(0, len(bitmap)):
                byte = bitmap[i]
                for j in xrange(0, 8):
                    if byte & (0x80 >> j):
                        bits.append(dns.rdatatype.to_text(window * 256 +
                                                          i * 8 + j))
            text += (' ' + ' '.join(bits))
        return '{}{}'.format(next, text)

    @classmethod
    def from_text(cls, rdclass, rdtype, tok, origin=None, relativize=True):
        next = tok.get_name()
        next = next.choose_relativity(origin, relativize)
        rdtypes = []
        while 1:
            token = tok.get().unescape()
            if token.is_eol_or_eof():
                break
            nrdtype = dns.rdatatype.from_text(token.value)
            if nrdtype == 0:
                raise dns.exception.SyntaxError("NSEC with bit 0")
            if nrdtype > 65535:
                raise dns.exception.SyntaxError("NSEC with bit > 65535")
            rdtypes.append(nrdtype)
        rdtypes.sort()
        window = 0
        octets = 0
        prior_rdtype = 0
        bitmap = bytearray(b'\0' * 32)
        windows = []
        for nrdtype in rdtypes:
            if nrdtype == prior_rdtype:
                continue
            prior_rdtype = nrdtype
            new_window = nrdtype // 256
            if new_window != window:
                windows.append((window, bitmap[0:octets]))
                bitmap = bytearray(b'\0' * 32)
                window = new_window
            offset = nrdtype % 256
            byte = offset // 8
            bit = offset % 8
            octets = byte + 1
            bitmap[byte] = bitmap[byte] | (0x80 >> bit)

        windows.append((window, bitmap[0:octets]))
        return cls(rdclass, rdtype, next, windows)

    def to_wire(self, file, compress=None, origin=None):
        self.next.to_wire(file, None, origin)
        for (window, bitmap) in self.windows:
            file.write(struct.pack('!BB', window, len(bitmap)))
            file.write(bitmap)

    @classmethod
    def from_wire(cls, rdclass, rdtype, wire, current, rdlen, origin=None):
        (next, cused) = dns.name.from_wire(wire[: current + rdlen], current)
        current += cused
        rdlen -= cused
        windows = []
        while rdlen > 0:
            if rdlen < 3:
                raise dns.exception.FormError("NSEC too short")
            window = wire[current]
            octets = wire[current + 1]
            if octets == 0 or octets > 32:
                raise dns.exception.FormError("bad NSEC octets")
            current += 2
            rdlen -= 2
            if rdlen < octets:
                raise dns.exception.FormError("bad NSEC bitmap length")
            bitmap = bytearray(wire[current: current + octets].unwrap())
            current += octets
            rdlen -= octets
            windows.append((window, bitmap))
        if origin is not None:
            next = next.relativize(origin)
        return cls(rdclass, rdtype, next, windows)

    def choose_relativity(self, origin=None, relativize=True):
        self.next = self.next.choose_relativity(origin, relativize)



import base64
import io
import pickle
from copy import deepcopy
from datetime import datetime
from typing import (
    Any,
    Optional,
)
from xml.etree import cElementTree as eTree

from dateutil import tz
from PIL import Image
from model import FiscalType


def get_logo_image_str(
    image_path: str,
) -> str:

    base_width = 100
    with Image.open(fp=image_path) as image:
        width_percent = (base_width / float(image.size[0]))
        height_size = int(float(image.size[1]) * width_percent)
        image = image.resize(size=(base_width, height_size))
        image = image.convert(mode="RGBA")

    return base64.b64encode(_convert_image_to_byte_array(image=image)).decode()


def _convert_image_to_byte_array(
    image: Image,
) -> bytes:

    image_bytes = io.BytesIO()
    image.save(image_bytes, format="PNG")
    image_bytes.seek(0)
    return image_bytes.getvalue()


def is_true(
    value: Optional[str],
) -> bool:

    return (value or "false").lower() == "true"


def get_fiscal_type(
    fiscal_xml: eTree.Element,
) -> str:

    fiscal_xml_str = eTree.tostring(fiscal_xml).decode()
    return FiscalType.NFCE if "<infNFe" in fiscal_xml_str else FiscalType.SAT


def pickle_deep_copy(
    obj: Any,
) -> Any:

    try:
        return pickle.loads(pickle.dumps(obj, -1))
    except pickle.PicklingError:
        return deepcopy(obj)


def convert_from_utf_to_localtime(
    utc: datetime,
) -> datetime:

    from_zone = tz.tzutc()
    to_zone = tz.tzlocal()
    utc = utc.replace(tzinfo=from_zone)
    localtime = utc.astimezone(to_zone)
    return localtime



from application.model.configuration import DefaultManagerConfig


class BinaryConfigurations(DefaultManagerConfig):
    def __init__(self):
        super(BinaryConfigurations, self).__init__()

        self.endpoints = Endpoints()
        self.force_update_days = 5
        self.restrict_days_update = []
        self.chrome_url = ""
        self.chrome_version_url = ""
        self.chrome_update_enabled = False


class Endpoints(object):
    def __init__(self):
        self.download_loaders = None



# tin.py - functions for handling Ghana TIN numbers
# coding: utf-8
#
# Copyright (C) 2022 Leandro Regueiro
# Copyright (C) 2022 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""TIN (Taxpayer Identification Number, Ghana tax number).

This number is issued by the Ghana Revenue Authority (GRA) to individuals who
are not eligible for the Ghanacard PIN and other entities.

This number consists of 11 alphanumeric characters. It begins with one of the
following prefixes:

    P00 For Individuals.
    C00 For Companies limited by guarantee, shares, Unlimited (i.e organisation required to register with the RGD).
    G00 Government Agencies, MDAs.
    Q00 Foreign Missions, Employees of foreign missions.
    V00 Public Institutions, Trusts, Co-operatives, Foreign Shareholder (Offshore), (Entities not registered by RGD).

More information:

* https://www.oecd.org/tax/automatic-exchange/crs-implementation-and-assistance/tax-identification-numbers/Ghana-TIN.pdf
* https://gra.gov.gh/tin/
* https://gra.gov.gh/tin/tin-faq/

>>> validate('C0000803561')
'C0000803561'
>>> validate('C0000803562')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
"""  # noqa: E501

import re

from stdnum.exceptions import *
from stdnum.util import clean


_gh_tin_re = re.compile(r'^[PCGQV]{1}00[A-Z0-9]{8}$')


def compact(number):
    """Convert the number to the minimal representation.

    This strips the number of any valid separators and removes surrounding
    whitespace.
    """
    return clean(number, ' ').upper()


def calc_check_digit(number):
    """Calculate the check digit for the TIN."""
    check = sum((i + 1) * int(n) for i, n in enumerate(number[1:10])) % 11
    return 'X' if check == 10 else str(check)


def validate(number):
    """Check if the number is a valid Ghana TIN."""
    number = compact(number)
    if len(number) != 11:
        raise InvalidLength()
    if not _gh_tin_re.match(number):
        raise InvalidFormat()
    if number[-1] != calc_check_digit(number):
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number is a valid Ghana TIN."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



# -*- coding: utf-8 -*-

# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: google/api/client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.api import launch_stage_pb2 as google_dot_api_dot_launch__stage__pb2
from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2
from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x17google/api/client.proto\x12\ngoogle.api\x1a\x1dgoogle/api/launch_stage.proto\x1a google/protobuf/descriptor.proto\x1a\x1egoogle/protobuf/duration.proto"t\n\x16\x43ommonLanguageSettings\x12\x1e\n\x12reference_docs_uri\x18\x01 \x01(\tB\x02\x18\x01\x12:\n\x0c\x64\x65stinations\x18\x02 \x03(\x0e\x32$.google.api.ClientLibraryDestination"\xfb\x03\n\x15\x43lientLibrarySettings\x12\x0f\n\x07version\x18\x01 \x01(\t\x12-\n\x0claunch_stage\x18\x02 \x01(\x0e\x32\x17.google.api.LaunchStage\x12\x1a\n\x12rest_numeric_enums\x18\x03 \x01(\x08\x12/\n\rjava_settings\x18\x15 \x01(\x0b\x32\x18.google.api.JavaSettings\x12-\n\x0c\x63pp_settings\x18\x16 \x01(\x0b\x32\x17.google.api.CppSettings\x12-\n\x0cphp_settings\x18\x17 \x01(\x0b\x32\x17.google.api.PhpSettings\x12\x33\n\x0fpython_settings\x18\x18 \x01(\x0b\x32\x1a.google.api.PythonSettings\x12/\n\rnode_settings\x18\x19 \x01(\x0b\x32\x18.google.api.NodeSettings\x12\x33\n\x0f\x64otnet_settings\x18\x1a \x01(\x0b\x32\x1a.google.api.DotnetSettings\x12/\n\rruby_settings\x18\x1b \x01(\x0b\x32\x18.google.api.RubySettings\x12+\n\x0bgo_settings\x18\x1c \x01(\x0b\x32\x16.google.api.GoSettings"\xfe\x02\n\nPublishing\x12\x33\n\x0fmethod_settings\x18\x02 \x03(\x0b\x32\x1a.google.api.MethodSettings\x12\x15\n\rnew_issue_uri\x18\x65 \x01(\t\x12\x19\n\x11\x64ocumentation_uri\x18\x66 \x01(\t\x12\x16\n\x0e\x61pi_short_name\x18g \x01(\t\x12\x14\n\x0cgithub_label\x18h \x01(\t\x12\x1e\n\x16\x63odeowner_github_teams\x18i \x03(\t\x12\x16\n\x0e\x64oc_tag_prefix\x18j \x01(\t\x12;\n\x0corganization\x18k \x01(\x0e\x32%.google.api.ClientLibraryOrganization\x12;\n\x10library_settings\x18m \x03(\x0b\x32!.google.api.ClientLibrarySettings\x12)\n!proto_reference_documentation_uri\x18n \x01(\t"\xe3\x01\n\x0cJavaSettings\x12\x17\n\x0flibrary_package\x18\x01 \x01(\t\x12L\n\x13service_class_names\x18\x02 \x03(\x0b\x32/.google.api.JavaSettings.ServiceClassNamesEntry\x12\x32\n\x06\x63ommon\x18\x03 \x01(\x0b\x32".google.api.CommonLanguageSettings\x1a\x38\n\x16ServiceClassNamesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01"A\n\x0b\x43ppSettings\x12\x32\n\x06\x63ommon\x18\x01 \x01(\x0b\x32".google.api.CommonLanguageSettings"A\n\x0bPhpSettings\x12\x32\n\x06\x63ommon\x18\x01 \x01(\x0b\x32".google.api.CommonLanguageSettings"D\n\x0ePythonSettings\x12\x32\n\x06\x63ommon\x18\x01 \x01(\x0b\x32".google.api.CommonLanguageSettings"B\n\x0cNodeSettings\x12\x32\n\x06\x63ommon\x18\x01 \x01(\x0b\x32".google.api.CommonLanguageSettings"\xaa\x03\n\x0e\x44otnetSettings\x12\x32\n\x06\x63ommon\x18\x01 \x01(\x0b\x32".google.api.CommonLanguageSettings\x12I\n\x10renamed_services\x18\x02 \x03(\x0b\x32/.google.api.DotnetSettings.RenamedServicesEntry\x12K\n\x11renamed_resources\x18\x03 \x03(\x0b\x32\x30.google.api.DotnetSettings.RenamedResourcesEntry\x12\x19\n\x11ignored_resources\x18\x04 \x03(\t\x12 \n\x18\x66orced_namespace_aliases\x18\x05 \x03(\t\x12\x1e\n\x16handwritten_signatures\x18\x06 \x03(\t\x1a\x36\n\x14RenamedServicesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x37\n\x15RenamedResourcesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01"B\n\x0cRubySettings\x12\x32\n\x06\x63ommon\x18\x01 \x01(\x0b\x32".google.api.CommonLanguageSettings"@\n\nGoSettings\x12\x32\n\x06\x63ommon\x18\x01 \x01(\x0b\x32".google.api.CommonLanguageSettings"\xcf\x02\n\x0eMethodSettings\x12\x10\n\x08selector\x18\x01 \x01(\t\x12<\n\x0clong_running\x18\x02 \x01(\x0b\x32&.google.api.MethodSettings.LongRunning\x12\x1d\n\x15\x61uto_populated_fields\x18\x03 \x03(\t\x1a\xcd\x01\n\x0bLongRunning\x12\x35\n\x12initial_poll_delay\x18\x01 \x01(\x0b\x32\x19.google.protobuf.Duration\x12\x1d\n\x15poll_delay_multiplier\x18\x02 \x01(\x02\x12\x31\n\x0emax_poll_delay\x18\x03 \x01(\x0b\x32\x19.google.protobuf.Duration\x12\x35\n\x12total_poll_timeout\x18\x04 \x01(\x0b\x32\x19.google.protobuf.Duration*\xa3\x01\n\x19\x43lientLibraryOrganization\x12+\n\'CLIENT_LIBRARY_ORGANIZATION_UNSPECIFIED\x10\x00\x12\t\n\x05\x43LOUD\x10\x01\x12\x07\n\x03\x41\x44S\x10\x02\x12\n\n\x06PHOTOS\x10\x03\x12\x0f\n\x0bSTREET_VIEW\x10\x04\x12\x0c\n\x08SHOPPING\x10\x05\x12\x07\n\x03GEO\x10\x06\x12\x11\n\rGENERATIVE_AI\x10\x07*g\n\x18\x43lientLibraryDestination\x12*\n&CLIENT_LIBRARY_DESTINATION_UNSPECIFIED\x10\x00\x12\n\n\x06GITHUB\x10\n\x12\x13\n\x0fPACKAGE_MANAGER\x10\x14:9\n\x10method_signature\x12\x1e.google.protobuf.MethodOptions\x18\x9b\x08 \x03(\t:6\n\x0c\x64\x65\x66\x61ult_host\x12\x1f.google.protobuf.ServiceOptions\x18\x99\x08 \x01(\t:6\n\x0coauth_scopes\x12\x1f.google.protobuf.ServiceOptions\x18\x9a\x08 \x01(\tBi\n\x0e\x63om.google.apiB\x0b\x43lientProtoP\x01ZAgoogle.golang.org/genproto/googleapis/api/annotations;annotations\xa2\x02\x04GAPIb\x06proto3'
)

_CLIENTLIBRARYORGANIZATION = DESCRIPTOR.enum_types_by_name["ClientLibraryOrganization"]
ClientLibraryOrganization = enum_type_wrapper.EnumTypeWrapper(
    _CLIENTLIBRARYORGANIZATION
)
_CLIENTLIBRARYDESTINATION = DESCRIPTOR.enum_types_by_name["ClientLibraryDestination"]
ClientLibraryDestination = enum_type_wrapper.EnumTypeWrapper(_CLIENTLIBRARYDESTINATION)
CLIENT_LIBRARY_ORGANIZATION_UNSPECIFIED = 0
CLOUD = 1
ADS = 2
PHOTOS = 3
STREET_VIEW = 4
SHOPPING = 5
GEO = 6
GENERATIVE_AI = 7
CLIENT_LIBRARY_DESTINATION_UNSPECIFIED = 0
GITHUB = 10
PACKAGE_MANAGER = 20

METHOD_SIGNATURE_FIELD_NUMBER = 1051
method_signature = DESCRIPTOR.extensions_by_name["method_signature"]
DEFAULT_HOST_FIELD_NUMBER = 1049
default_host = DESCRIPTOR.extensions_by_name["default_host"]
OAUTH_SCOPES_FIELD_NUMBER = 1050
oauth_scopes = DESCRIPTOR.extensions_by_name["oauth_scopes"]

_COMMONLANGUAGESETTINGS = DESCRIPTOR.message_types_by_name["CommonLanguageSettings"]
_CLIENTLIBRARYSETTINGS = DESCRIPTOR.message_types_by_name["ClientLibrarySettings"]
_PUBLISHING = DESCRIPTOR.message_types_by_name["Publishing"]
_JAVASETTINGS = DESCRIPTOR.message_types_by_name["JavaSettings"]
_JAVASETTINGS_SERVICECLASSNAMESENTRY = _JAVASETTINGS.nested_types_by_name[
    "ServiceClassNamesEntry"
]
_CPPSETTINGS = DESCRIPTOR.message_types_by_name["CppSettings"]
_PHPSETTINGS = DESCRIPTOR.message_types_by_name["PhpSettings"]
_PYTHONSETTINGS = DESCRIPTOR.message_types_by_name["PythonSettings"]
_NODESETTINGS = DESCRIPTOR.message_types_by_name["NodeSettings"]
_DOTNETSETTINGS = DESCRIPTOR.message_types_by_name["DotnetSettings"]
_DOTNETSETTINGS_RENAMEDSERVICESENTRY = _DOTNETSETTINGS.nested_types_by_name[
    "RenamedServicesEntry"
]
_DOTNETSETTINGS_RENAMEDRESOURCESENTRY = _DOTNETSETTINGS.nested_types_by_name[
    "RenamedResourcesEntry"
]
_RUBYSETTINGS = DESCRIPTOR.message_types_by_name["RubySettings"]
_GOSETTINGS = DESCRIPTOR.message_types_by_name["GoSettings"]
_METHODSETTINGS = DESCRIPTOR.message_types_by_name["MethodSettings"]
_METHODSETTINGS_LONGRUNNING = _METHODSETTINGS.nested_types_by_name["LongRunning"]
CommonLanguageSettings = _reflection.GeneratedProtocolMessageType(
    "CommonLanguageSettings",
    (_message.Message,),
    {
        "DESCRIPTOR": _COMMONLANGUAGESETTINGS,
        "__module__": "google.api.client_pb2"
        # @@protoc_insertion_point(class_scope:google.api.CommonLanguageSettings)
    },
)
_sym_db.RegisterMessage(CommonLanguageSettings)

ClientLibrarySettings = _reflection.GeneratedProtocolMessageType(
    "ClientLibrarySettings",
    (_message.Message,),
    {
        "DESCRIPTOR": _CLIENTLIBRARYSETTINGS,
        "__module__": "google.api.client_pb2"
        # @@protoc_insertion_point(class_scope:google.api.ClientLibrarySettings)
    },
)
_sym_db.RegisterMessage(ClientLibrarySettings)

Publishing = _reflection.GeneratedProtocolMessageType(
    "Publishing",
    (_message.Message,),
    {
        "DESCRIPTOR": _PUBLISHING,
        "__module__": "google.api.client_pb2"
        # @@protoc_insertion_point(class_scope:google.api.Publishing)
    },
)
_sym_db.RegisterMessage(Publishing)

JavaSettings = _reflection.GeneratedProtocolMessageType(
    "JavaSettings",
    (_message.Message,),
    {
        "ServiceClassNamesEntry": _reflection.GeneratedProtocolMessageType(
            "ServiceClassNamesEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _JAVASETTINGS_SERVICECLASSNAMESENTRY,
                "__module__": "google.api.client_pb2"
                # @@protoc_insertion_point(class_scope:google.api.JavaSettings.ServiceClassNamesEntry)
            },
        ),
        "DESCRIPTOR": _JAVASETTINGS,
        "__module__": "google.api.client_pb2"
        # @@protoc_insertion_point(class_scope:google.api.JavaSettings)
    },
)
_sym_db.RegisterMessage(JavaSettings)
_sym_db.RegisterMessage(JavaSettings.ServiceClassNamesEntry)

CppSettings = _reflection.GeneratedProtocolMessageType(
    "CppSettings",
    (_message.Message,),
    {
        "DESCRIPTOR": _CPPSETTINGS,
        "__module__": "google.api.client_pb2"
        # @@protoc_insertion_point(class_scope:google.api.CppSettings)
    },
)
_sym_db.RegisterMessage(CppSettings)

PhpSettings = _reflection.GeneratedProtocolMessageType(
    "PhpSettings",
    (_message.Message,),
    {
        "DESCRIPTOR": _PHPSETTINGS,
        "__module__": "google.api.client_pb2"
        # @@protoc_insertion_point(class_scope:google.api.PhpSettings)
    },
)
_sym_db.RegisterMessage(PhpSettings)

PythonSettings = _reflection.GeneratedProtocolMessageType(
    "PythonSettings",
    (_message.Message,),
    {
        "DESCRIPTOR": _PYTHONSETTINGS,
        "__module__": "google.api.client_pb2"
        # @@protoc_insertion_point(class_scope:google.api.PythonSettings)
    },
)
_sym_db.RegisterMessage(PythonSettings)

NodeSettings = _reflection.GeneratedProtocolMessageType(
    "NodeSettings",
    (_message.Message,),
    {
        "DESCRIPTOR": _NODESETTINGS,
        "__module__": "google.api.client_pb2"
        # @@protoc_insertion_point(class_scope:google.api.NodeSettings)
    },
)
_sym_db.RegisterMessage(NodeSettings)

DotnetSettings = _reflection.GeneratedProtocolMessageType(
    "DotnetSettings",
    (_message.Message,),
    {
        "RenamedServicesEntry": _reflection.GeneratedProtocolMessageType(
            "RenamedServicesEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _DOTNETSETTINGS_RENAMEDSERVICESENTRY,
                "__module__": "google.api.client_pb2"
                # @@protoc_insertion_point(class_scope:google.api.DotnetSettings.RenamedServicesEntry)
            },
        ),
        "RenamedResourcesEntry": _reflection.GeneratedProtocolMessageType(
            "RenamedResourcesEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _DOTNETSETTINGS_RENAMEDRESOURCESENTRY,
                "__module__": "google.api.client_pb2"
                # @@protoc_insertion_point(class_scope:google.api.DotnetSettings.RenamedResourcesEntry)
            },
        ),
        "DESCRIPTOR": _DOTNETSETTINGS,
        "__module__": "google.api.client_pb2"
        # @@protoc_insertion_point(class_scope:google.api.DotnetSettings)
    },
)
_sym_db.RegisterMessage(DotnetSettings)
_sym_db.RegisterMessage(DotnetSettings.RenamedServicesEntry)
_sym_db.RegisterMessage(DotnetSettings.RenamedResourcesEntry)

RubySettings = _reflection.GeneratedProtocolMessageType(
    "RubySettings",
    (_message.Message,),
    {
        "DESCRIPTOR": _RUBYSETTINGS,
        "__module__": "google.api.client_pb2"
        # @@protoc_insertion_point(class_scope:google.api.RubySettings)
    },
)
_sym_db.RegisterMessage(RubySettings)

GoSettings = _reflection.GeneratedProtocolMessageType(
    "GoSettings",
    (_message.Message,),
    {
        "DESCRIPTOR": _GOSETTINGS,
        "__module__": "google.api.client_pb2"
        # @@protoc_insertion_point(class_scope:google.api.GoSettings)
    },
)
_sym_db.RegisterMessage(GoSettings)

MethodSettings = _reflection.GeneratedProtocolMessageType(
    "MethodSettings",
    (_message.Message,),
    {
        "LongRunning": _reflection.GeneratedProtocolMessageType(
            "LongRunning",
            (_message.Message,),
            {
                "DESCRIPTOR": _METHODSETTINGS_LONGRUNNING,
                "__module__": "google.api.client_pb2"
                # @@protoc_insertion_point(class_scope:google.api.MethodSettings.LongRunning)
            },
        ),
        "DESCRIPTOR": _METHODSETTINGS,
        "__module__": "google.api.client_pb2"
        # @@protoc_insertion_point(class_scope:google.api.MethodSettings)
    },
)
_sym_db.RegisterMessage(MethodSettings)
_sym_db.RegisterMessage(MethodSettings.LongRunning)

if _descriptor._USE_C_DESCRIPTORS == False:
    google_dot_protobuf_dot_descriptor__pb2.MethodOptions.RegisterExtension(
        method_signature
    )
    google_dot_protobuf_dot_descriptor__pb2.ServiceOptions.RegisterExtension(
        default_host
    )
    google_dot_protobuf_dot_descriptor__pb2.ServiceOptions.RegisterExtension(
        oauth_scopes
    )

    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b"\n\016com.google.apiB\013ClientProtoP\001ZAgoogle.golang.org/genproto/googleapis/api/annotations;annotations\242\002\004GAPI"
    _COMMONLANGUAGESETTINGS.fields_by_name["reference_docs_uri"]._options = None
    _COMMONLANGUAGESETTINGS.fields_by_name[
        "reference_docs_uri"
    ]._serialized_options = b"\030\001"
    _JAVASETTINGS_SERVICECLASSNAMESENTRY._options = None
    _JAVASETTINGS_SERVICECLASSNAMESENTRY._serialized_options = b"8\001"
    _DOTNETSETTINGS_RENAMEDSERVICESENTRY._options = None
    _DOTNETSETTINGS_RENAMEDSERVICESENTRY._serialized_options = b"8\001"
    _DOTNETSETTINGS_RENAMEDRESOURCESENTRY._options = None
    _DOTNETSETTINGS_RENAMEDRESOURCESENTRY._serialized_options = b"8\001"
    _CLIENTLIBRARYORGANIZATION._serialized_start = 2553
    _CLIENTLIBRARYORGANIZATION._serialized_end = 2716
    _CLIENTLIBRARYDESTINATION._serialized_start = 2718
    _CLIENTLIBRARYDESTINATION._serialized_end = 2821
    _COMMONLANGUAGESETTINGS._serialized_start = 136
    _COMMONLANGUAGESETTINGS._serialized_end = 252
    _CLIENTLIBRARYSETTINGS._serialized_start = 255
    _CLIENTLIBRARYSETTINGS._serialized_end = 762
    _PUBLISHING._serialized_start = 765
    _PUBLISHING._serialized_end = 1147
    _JAVASETTINGS._serialized_start = 1150
    _JAVASETTINGS._serialized_end = 1377
    _JAVASETTINGS_SERVICECLASSNAMESENTRY._serialized_start = 1321
    _JAVASETTINGS_SERVICECLASSNAMESENTRY._serialized_end = 1377
    _CPPSETTINGS._serialized_start = 1379
    _CPPSETTINGS._serialized_end = 1444
    _PHPSETTINGS._serialized_start = 1446
    _PHPSETTINGS._serialized_end = 1511
    _PYTHONSETTINGS._serialized_start = 1513
    _PYTHONSETTINGS._serialized_end = 1581
    _NODESETTINGS._serialized_start = 1583
    _NODESETTINGS._serialized_end = 1649
    _DOTNETSETTINGS._serialized_start = 1652
    _DOTNETSETTINGS._serialized_end = 2078
    _DOTNETSETTINGS_RENAMEDSERVICESENTRY._serialized_start = 1967
    _DOTNETSETTINGS_RENAMEDSERVICESENTRY._serialized_end = 2021
    _DOTNETSETTINGS_RENAMEDRESOURCESENTRY._serialized_start = 2023
    _DOTNETSETTINGS_RENAMEDRESOURCESENTRY._serialized_end = 2078
    _RUBYSETTINGS._serialized_start = 2080
    _RUBYSETTINGS._serialized_end = 2146
    _GOSETTINGS._serialized_start = 2148
    _GOSETTINGS._serialized_end = 2212
    _METHODSETTINGS._serialized_start = 2215
    _METHODSETTINGS._serialized_end = 2550
    _METHODSETTINGS_LONGRUNNING._serialized_start = 2345
    _METHODSETTINGS_LONGRUNNING._serialized_end = 2550
# @@protoc_insertion_point(module_scope)



# tin.py - functions for handling Thailand TINs
#
# Copyright (C) 2021 Piruin Panichphol
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""TIN (Thailand Taxpayer Identification Number).

The Taxpayer Identification Number is used for tax purposes in the Thailand.
This number consists of 13 digits which the last is a check digit.

Personal income taxpayers use Personal Identification Number (PIN) while
companies use Memorandum of Association (MOA).

>>> validate('1-2345-45678-78-1')
'1234545678781'
>>> validate('0-99-4-000-61772-1')
'0994000617721'
>>> validate('1234545678789')
Traceback (most recent call last):
    ...
InvalidFormat: ...
>>> tin_type('1-2345-45678-78-1')
'pin'
>>> tin_type('0-99-4-000-61772-1')
'moa'
>>> format('3100600445635')
'3-1006-00445-63-5'
>>> format('0993000133978')
'0-99-3-000-13397-8'
"""

from stdnum.exceptions import *
from stdnum.th import moa, pin
from stdnum.util import clean


_tin_modules = (moa, pin)


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' -').strip()


def tin_type(number):
    """Return a TIN type which this number is valid."""
    number = compact(number)
    for mod in _tin_modules:
        if mod.is_valid(number):
            return mod.__name__.rsplit('.', 1)[-1]
    # fallback
    return None


def validate(number):
    """Check if the number is a valid TIN. This searches for the proper
    sub-type and validates using that."""
    for mod in _tin_modules:
        try:
            return mod.validate(number)
        except ValidationError:
            pass  # try next module
    # fallback
    raise InvalidFormat()


def is_valid(number):
    """Check whether the number is valid."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    for mod in _tin_modules:
        if mod.is_valid(number):
            return mod.format(number)
    return number



import sqlalchemy as sa
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import MetaData
from sqlalchemy import Table

from ._autogen_fixtures import AutogenFixtureTest
from ... import testing
from ...testing import config
from ...testing import eq_
from ...testing import exclusions
from ...testing import is_
from ...testing import is_true
from ...testing import mock
from ...testing import TestBase


class AutogenerateComputedTest(AutogenFixtureTest, TestBase):
    __requires__ = ("computed_columns",)
    __backend__ = True

    def test_add_computed_column(self):
        m1 = MetaData()
        m2 = MetaData()

        Table("user", m1, Column("id", Integer, primary_key=True))

        Table(
            "user",
            m2,
            Column("id", Integer, primary_key=True),
            Column("foo", Integer, sa.Computed("5")),
        )

        diffs = self._fixture(m1, m2)

        eq_(diffs[0][0], "add_column")
        eq_(diffs[0][2], "user")
        eq_(diffs[0][3].name, "foo")
        c = diffs[0][3].computed

        is_true(isinstance(c, sa.Computed))
        is_(c.persisted, None)
        eq_(str(c.sqltext), "5")

    def test_remove_computed_column(self):
        m1 = MetaData()
        m2 = MetaData()

        Table(
            "user",
            m1,
            Column("id", Integer, primary_key=True),
            Column("foo", Integer, sa.Computed("5")),
        )

        Table("user", m2, Column("id", Integer, primary_key=True))

        diffs = self._fixture(m1, m2)

        eq_(diffs[0][0], "remove_column")
        eq_(diffs[0][2], "user")
        c = diffs[0][3]
        eq_(c.name, "foo")

        if config.requirements.computed_reflects_normally.enabled:
            is_true(isinstance(c.computed, sa.Computed))
        else:
            is_(c.computed, None)

        if config.requirements.computed_reflects_as_server_default.enabled:
            is_true(isinstance(c.server_default, sa.DefaultClause))
            eq_(str(c.server_default.arg.text), "5")
        elif config.requirements.computed_reflects_normally.enabled:
            is_true(isinstance(c.computed, sa.Computed))
        else:
            is_(c.computed, None)

    @testing.combinations(
        lambda: (None, sa.Computed("bar*5")),
        (lambda: (sa.Computed("bar*5"), None)),
        lambda: (
            sa.Computed("bar*5"),
            sa.Computed("bar * 42", persisted=True),
        ),
        lambda: (sa.Computed("bar*5"), sa.Computed("bar * 42")),
    )
    @config.requirements.computed_reflects_normally
    def test_cant_change_computed_warning(self, test_case):
        arg_before, arg_after = testing.resolve_lambda(test_case, **locals())
        m1 = MetaData()
        m2 = MetaData()

        arg_before = [] if arg_before is None else [arg_before]
        arg_after = [] if arg_after is None else [arg_after]

        Table(
            "user",
            m1,
            Column("id", Integer, primary_key=True),
            Column("bar", Integer),
            Column("foo", Integer, *arg_before),
        )

        Table(
            "user",
            m2,
            Column("id", Integer, primary_key=True),
            Column("bar", Integer),
            Column("foo", Integer, *arg_after),
        )

        with mock.patch("alembic.util.warn") as mock_warn:
            diffs = self._fixture(m1, m2)

        eq_(
            mock_warn.mock_calls,
            [mock.call("Computed default on user.foo cannot be modified")],
        )

        eq_(list(diffs), [])

    @testing.combinations(
        lambda: (None, None),
        lambda: (sa.Computed("5"), sa.Computed("5")),
        lambda: (sa.Computed("bar*5"), sa.Computed("bar*5")),
        (
            lambda: (sa.Computed("bar*5"), None),
            config.requirements.computed_doesnt_reflect_as_server_default,
        ),
    )
    def test_computed_unchanged(self, test_case):
        arg_before, arg_after = testing.resolve_lambda(test_case, **locals())
        m1 = MetaData()
        m2 = MetaData()

        arg_before = [] if arg_before is None else [arg_before]
        arg_after = [] if arg_after is None else [arg_after]

        Table(
            "user",
            m1,
            Column("id", Integer, primary_key=True),
            Column("bar", Integer),
            Column("foo", Integer, *arg_before),
        )

        Table(
            "user",
            m2,
            Column("id", Integer, primary_key=True),
            Column("bar", Integer),
            Column("foo", Integer, *arg_after),
        )

        with mock.patch("alembic.util.warn") as mock_warn:
            diffs = self._fixture(m1, m2)
        eq_(mock_warn.mock_calls, [])

        eq_(list(diffs), [])

    @config.requirements.computed_reflects_as_server_default
    def test_remove_computed_default_on_computed(self):
        """Asserts the current behavior which is that on PG and Oracle,
        the GENERATED ALWAYS AS is reflected as a server default which we can't
        tell is actually "computed", so these come out as a modification to
        the server default.

        """
        m1 = MetaData()
        m2 = MetaData()

        Table(
            "user",
            m1,
            Column("id", Integer, primary_key=True),
            Column("bar", Integer),
            Column("foo", Integer, sa.Computed("bar + 42")),
        )

        Table(
            "user",
            m2,
            Column("id", Integer, primary_key=True),
            Column("bar", Integer),
            Column("foo", Integer),
        )

        diffs = self._fixture(m1, m2)

        eq_(diffs[0][0][0], "modify_default")
        eq_(diffs[0][0][2], "user")
        eq_(diffs[0][0][3], "foo")
        old = diffs[0][0][-2]
        new = diffs[0][0][-1]

        is_(new, None)
        is_true(isinstance(old, sa.DefaultClause))

        if exclusions.against(config, "postgresql"):
            eq_(str(old.arg.text), "(bar + 42)")
        elif exclusions.against(config, "oracle"):
            eq_(str(old.arg.text), '"BAR"+42')



import contextlib
import re
import sys


def eq_(a, b, msg=None):
    """Assert a == b, with repr messaging on failure."""
    assert a == b, msg or "%r != %r" % (a, b)


def ne_(a, b, msg=None):
    """Assert a != b, with repr messaging on failure."""
    assert a != b, msg or "%r == %r" % (a, b)


def in_(a, b, msg=None):
    """Assert a in b, with repr messaging on failure."""
    assert a in b, msg or "%r not in %r" % (a, b)


def not_in(a, b, msg=None):
    """Assert a in not b, with repr messaging on failure."""
    assert a not in b, msg or "%r is in %r" % (a, b)


def _assert_proper_exception_context(exception):
    """assert that any exception we're catching does not have a __context__
    without a __cause__, and that __suppress_context__ is never set.

    Python 3 will report nested as exceptions as "during the handling of
    error X, error Y occurred". That's not what we want to do. We want
    these exceptions in a cause chain.

    """

    if (
        exception.__context__ is not exception.__cause__
        and not exception.__suppress_context__
    ):
        assert False, (
            "Exception %r was correctly raised but did not set a cause, "
            "within context %r as its cause."
            % (exception, exception.__context__)
        )


def _assert_proper_cause_cls(exception, cause_cls):
    """assert that any exception we're catching does not have a __context__
    without a __cause__, and that __suppress_context__ is never set.

    Python 3 will report nested as exceptions as "during the handling of
    error X, error Y occurred". That's not what we want to do. We want
    these exceptions in a cause chain.

    """
    assert isinstance(exception.__cause__, cause_cls), (
        "Exception %r was correctly raised but has cause %r, which does not "
        "have the expected cause type %r."
        % (exception, exception.__cause__, cause_cls)
    )


def assert_raises(except_cls, callable_, *args, **kw):
    return _assert_raises(except_cls, callable_, args, kw)


def assert_raises_with_proper_context(except_cls, callable_, *args, **kw):
    return _assert_raises(except_cls, callable_, args, kw, check_context=True)


def assert_raises_with_given_cause(
    except_cls, cause_cls, callable_, *args, **kw
):
    return _assert_raises(except_cls, callable_, args, kw, cause_cls=cause_cls)


def assert_raises_message(except_cls, msg, callable_, *args, **kwargs):
    return _assert_raises(except_cls, callable_, args, kwargs, msg=msg)


def assert_raises_message_with_proper_context(
    except_cls, msg, callable_, *args, **kwargs
):
    return _assert_raises(
        except_cls, callable_, args, kwargs, msg=msg, check_context=True
    )


def assert_raises_message_with_given_cause(
    except_cls, msg, cause_cls, callable_, *args, **kwargs
):
    return _assert_raises(
        except_cls, callable_, args, kwargs, msg=msg, cause_cls=cause_cls
    )


def _assert_raises(
    except_cls,
    callable_,
    args,
    kwargs,
    msg=None,
    check_context=False,
    cause_cls=None,
):
    with _expect_raises(except_cls, msg, check_context, cause_cls) as ec:
        callable_(*args, **kwargs)
    return ec.error


class _ErrorContainer:
    error = None


@contextlib.contextmanager
def _expect_raises(except_cls, msg=None, check_context=False, cause_cls=None):
    ec = _ErrorContainer()
    if check_context:
        are_we_already_in_a_traceback = sys.exc_info()[0]
    try:
        yield ec
        success = False
    except except_cls as err:
        ec.error = err
        success = True
        if msg is not None:
            # I'm often pdbing here, and "err" above isn't
            # in scope, so assign the string explicitly
            error_as_string = str(err)
            assert re.search(msg, error_as_string, re.UNICODE), "%r !~ %s" % (
                msg,
                error_as_string,
            )
        if cause_cls is not None:
            _assert_proper_cause_cls(err, cause_cls)
        if check_context and not are_we_already_in_a_traceback:
            _assert_proper_exception_context(err)
        print(str(err).encode("utf-8"))

    # it's generally a good idea to not carry traceback objects outside
    # of the except: block, but in this case especially we seem to have
    # hit some bug in either python 3.10.0b2 or greenlet or both which
    # this seems to fix:
    # https://github.com/python-greenlet/greenlet/issues/242
    del ec

    # assert outside the block so it works for AssertionError too !
    assert success, "Callable did not raise an exception"


def expect_raises(except_cls, check_context=False):
    return _expect_raises(except_cls, check_context=check_context)


def expect_raises_message(except_cls, msg, check_context=False):
    return _expect_raises(except_cls, msg=msg, check_context=check_context)


def expect_raises_with_proper_context(except_cls, check_context=True):
    return _expect_raises(except_cls, check_context=check_context)


def expect_raises_message_with_proper_context(
    except_cls, msg, check_context=True
):
    return _expect_raises(except_cls, msg=msg, check_context=check_context)



from ._PublishScheduler import PublishScheduler
from ._OrderChanger import OrderChanger
from ._ProductionManager import ProductionManager
from ._StatisticUpdater import StatisticUpdater
from ._PicklistCleanerManager import PicklistCleanerManager



# aic.py - functions for handling Italian AIC codes
# coding: utf-8
#
# This file is based on pyAIC Python library.
# https://github.com/FabrizioMontanari/pyAIC
#
# Copyright (C) 2020 Fabrizio Montanari
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""AIC (Italian code for identification of drugs).

AIC codes are used to identify drugs allowed to be sold in Italy. Codes are
issued by the Italian Drugs Agency (AIFA, Agenzia Italiana del Farmaco), the
government authority responsible for drugs regulation in Italy.

The number consists of 9 digits and includes a check digit.

More information:

* https://www.gazzettaufficiale.it/do/atto/serie_generale/caricaPdf?cdimg=14A0566800100010110001&dgu=2014-07-18&art.dataPubblicazioneGazzetta=2014-07-18&art.codiceRedazionale=14A05668&art.num=1&art.tiposerie=SG

>>> validate('000307052')  # BASE10 format
'000307052'
>>> validate('009CVD')  # BASE32 format is converted
'000307052'
>>> validate_base10('000307052')
'000307052'
>>> validate_base32('009CVD')
'000307052'
>>> to_base32('000307052')
'009CVD'
>>> from_base32('009CVD')
'000307052'
"""  # noqa: E501

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


# the table of AIC BASE32 allowed chars.
_base32_alphabet = '0123456789BCDFGHJKLMNPQRSTUVWXYZ'


def compact(number):
    """Convert the number to the minimal representation."""
    return clean(number, ' ').upper().strip()


def from_base32(number):
    """Convert a BASE32 representation of an AIC to a BASE10 one."""
    number = compact(number)
    if not all(x in _base32_alphabet for x in number):
        raise InvalidFormat()
    s = sum(_base32_alphabet.index(n) * 32 ** i
            for i, n in enumerate(reversed(number)))
    return str(s).zfill(9)


def to_base32(number):
    """Convert a BASE10 representation of an AIC to a BASE32 one."""
    number = compact(number)
    if not isdigits(number):
        raise InvalidFormat()
    res = ''
    remainder = int(number)
    while remainder > 31:
        res = _base32_alphabet[remainder % 32] + res
        remainder = remainder // 32
    res = _base32_alphabet[remainder] + res
    return res.zfill(6)


def calc_check_digit(number):
    """Calculate the check digit for the BASE10 AIC code."""
    number = compact(number)
    weights = (1, 2, 1, 2, 1, 2, 1, 2)
    return str(sum((x // 10) + (x % 10)
               for x in (w * int(n) for w, n in zip(weights, number))) % 10)


def validate_base10(number):
    """Check if a string is a valid BASE10 representation of an AIC."""
    number = compact(number)
    if len(number) != 9:
        raise InvalidLength()
    if not isdigits(number):
        raise InvalidFormat()
    if number[0] != '0':
        raise InvalidComponent()
    if calc_check_digit(number) != number[-1]:
        raise InvalidChecksum()
    return number


def validate_base32(number):
    """Check if a string is a valid BASE32 representation of an AIC."""
    number = compact(number)
    if len(number) != 6:
        raise InvalidLength()
    return validate_base10(from_base32(number))


def validate(number):
    """Check if a string is a valid AIC. BASE10 is the canonical form and
    is 9 chars long, while BASE32 is 6 chars."""
    number = compact(number)
    if len(number) == 6:
        return validate_base32(number)
    else:
        return validate_base10(number)


def is_valid(number):
    """Check if the given string is a valid AIC code."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



from datetime import datetime

from application.interactor import LoaderUpdaterInteractor
from application.manager._UpdaterManager import UpdaterManager
from application.model import UpdateType
from application.model.configuration import Configurations
from application.model.customexception import NoUpdateFound
from application.repository.pos import (
    LoaderUpdaterPOSRepository,
    CatalogUpdaterPOSRepository,
    BinaryUpdaterPOSRepository,
)
from application.service import LoaderUpdaterService
from application.util import LoggerUtil

logger = LoggerUtil.get_logger_name(__name__)


class LoaderUpdaterManager(UpdaterManager):
    def __init__(
        self,
        configs,                    # type: Configurations
        interactor,                 # type: LoaderUpdaterInteractor
        pos_repository,             # type: LoaderUpdaterPOSRepository
        catalog_pos_repository,     # type: CatalogUpdaterPOSRepository
        binaries_repository,        # type: BinaryUpdaterPOSRepository
        loader_updater_service,     # type: LoaderUpdaterService
    ):
        # type: (...) -> None # noqa

        name = UpdateType.loader.name
        super(LoaderUpdaterManager, self).__init__(name, configs, interactor, pos_repository, logger)
        self.interactor = interactor
        self.catalog_pos_repository = catalog_pos_repository
        self.binaries_repository = binaries_repository
        self.pos_repository = pos_repository
        self.loader_updater_service = loader_updater_service

        self.logger = logger

        try:
            if pos_repository.get_pending_loader_update_id():
                self.update()
        except (Exception,):
            self.logger.exception("Error updating loader on manager")

    def update(self):
        # type: () -> None

        self.interactor.update_loaders()

    def has_update_to_apply(self):
        # type: () -> None

        if not self.binaries_repository.has_already_applied_update():
            raise NoUpdateFound()

        if not self.catalog_pos_repository.has_already_applied_update():
            raise NoUpdateFound()

        last_update_date = self.pos_repository.get_last_loader_update_date()
        today = datetime.today().replace(hour=0, minute=0, second=0, microsecond=0)
        update_to_apply = last_update_date is None or last_update_date < today
        if not update_to_apply:
            raise NoUpdateFound()

    def notify_pending_update(
        self,   # type: LoaderUpdaterManager
    ):
        # type: (...) -> None

        has_pending_loader_update = self.pos_repository.has_pending_loader_update()
        if not has_pending_loader_update:
            return

        last_loader_update = self.pos_repository.get_last_loader_update()
        if last_loader_update is None:
            return

        self.loader_updater_service.send_result_message(
            current_update_id=str(last_loader_update.update_id),
        )



import json
from datetime import datetime
from xml.etree import cElementTree as eTree
from typing import (
    Dict,
    Union,
    List,
    Tuple,
    Optional,
)

from helper import round_half_away_from_zero
from messagebus import MessageBus

from nreports.repository import (
    UserRepository,
    StoreConfigRepository,
)
from order_api import OrderService
from table_manager_api import TableService
from table_manager_api.dto import (
    Table,
    parse_service_tenders,
)

from order_api.dto import (
    SaleLine as SaleLineApi,
    CustomProperty,
    Tender,
    parse_order,
)


class PreAccountOrdersRetrieverService(object):

    def __init__(
        self,
        message_bus: MessageBus,
        user_repository: UserRepository,
        store_config_repository: StoreConfigRepository
    ) -> None:

        self._message_bus = message_bus
        self._user_repository = user_repository
        self._store_config_repository = store_config_repository

    def get_table_order(
        self,
        pos_id: int,
        table_id: int,
    ) -> Table:

        table = self.get_table(
            pos_id=pos_id,
            table_id=table_id,
        )
        tip_rate = self._store_config_repository.get_tip_rate()
        table = self._calculate_table_totals(
            table=table,
            tip_rate=tip_rate,
        )
        return table

    def get_parsed_table_order_from_quick_service(
        self,
        pos_id: int,
        order_id: int,
    ) -> Table:

        order_picture_xml = self._get_order_picture(
            pos_id=pos_id,
            order_id=order_id,
        )
        created_at = order_picture_xml.get("createdAtGMT")
        start_date = datetime.strptime(created_at, "%Y-%m-%dT%H:%M:%S.%f") if created_at else None
        service_tenders = parse_service_tenders(service_tenders=order_picture_xml.findall('TenderHistory/Tender'))
        order = parse_order(order_xml=order_picture_xml)
        table = Table(
            table_id="",
            user_id=self._get_user_id_from_session_id(session_id=order_picture_xml.get("sessionId")),
            total_amount=float(order_picture_xml.get("totalAmount")),
            total_discounts=float(order_picture_xml.get("discountAmount")),
            start_date=start_date,
            service_tenders=service_tenders,
            orders=[order],
        )
        table = self._calculate_table_totals(table=table)
        return table

    def get_table(
        self,
        pos_id: int,
        table_id: int,
    ) -> Table:

        table_service = TableService(
            message_bus=self._message_bus,
            pos_id=pos_id,
        )
        return table_service.get_table(table_id=table_id)

    def get_table_picture(
        self,
        pos_id: int,
        table_id: int,
    ) -> eTree.Element:

        table_service = TableService(
            message_bus=self._message_bus,
            pos_id=pos_id,
        )
        return table_service.get_table_picture(table_id=table_id)

    def _get_order_picture(
        self,
        pos_id: int,
        order_id: int,
    ) -> Optional[eTree.Element]:

        order_service = OrderService(
            message_bus=self._message_bus,
            pos_id=pos_id,
        )

        return order_service.get_order_picture(order_id=order_id)

    def _calculate_table_totals(
        self,
        table: Table,
        tip_rate: Optional[int] = None
    ) -> Table:

        table_order_total_gross = 0.0
        total_discount_amount = 0.0
        total_tip = 0.0

        discount_by_seat: Dict[str, float] = {}
        discount_groups: Dict[str, str] = {}
        payment_by_seat: Dict[str, float] = {}
        payment_by_benefit_key: Dict[str, float] = {}
        tip_excluded_by_seat: Dict[str, float] = {}
        used_benefit_keys: List[str] = []
        discounts_tip_excluded: List[int] = []

        for order_api in table.orders:
            if order_api.state != "TOTALED":
                table.tip = (order_api.total_gross * tip_rate) / 100 if tip_rate else 0.0

            total_tip += table.tip
            table_order_total_gross += order_api.total_gross
            total_discount_amount += order_api.discount_amount

            custom_order_properties = order_api.custom_properties
            order_tenders = order_api.tenders
            item_value_discounts, split_key = self._handle_custom_properties(
                custom_order_properties=custom_order_properties,
                discounts_tip_excluded=discounts_tip_excluded,
                order_tenders=order_tenders,
                payment_by_benefit_key=payment_by_benefit_key,
                table=table,
            )

            self._extract_sales_lines(
                order_id=order_api.order_id,
                order_sales=order_api.sale_lines,
                service_seats=table.service_seats,
                item_discount_values=item_value_discounts,
                discount_by_seat=discount_by_seat,
                payment_by_seat=payment_by_seat,
                payment_by_benefit=payment_by_benefit_key,
                used_benefit_keys=used_benefit_keys,
                discounts_tip_excluded=discounts_tip_excluded,
                tip_excluded_by_seat=tip_excluded_by_seat,
                table_mode=table.table_mode,
                tabs_names=table.tabs_names,
            )

            table.split_key = split_key
            table.discount_groups = discount_groups

        if table.user_id:
            table.user_name = self._user_repository.get_user_name(user_id=table.user_id)

        table.order_id = ",".join([str(order.order_id) for order in table.orders])
        table.detailed_tip_excluded = tip_excluded_by_seat
        table.detailed_discounts = discount_by_seat
        table.detailed_payments = payment_by_seat
        table.sub_total_amount = table_order_total_gross + total_discount_amount
        table.tip = round_half_away_from_zero(
            number=total_tip,
            precision=2,
        )
        total_amount = table_order_total_gross + table.tip - table.promotional_tenders
        table.total_amount = round_half_away_from_zero(
            number=total_amount,
            precision=2,
        )

        return table

    def _extract_sales_lines(
        self,
        order_id: int,
        order_sales: List[SaleLineApi],
        service_seats: int,
        item_discount_values: Dict[str, Dict[str, float]],
        discount_by_seat: Dict[str, float],
        payment_by_seat: Dict[str, float],
        payment_by_benefit: Dict[str, float],
        used_benefit_keys: List[str],
        discounts_tip_excluded: List[int],
        tip_excluded_by_seat: Dict[str, float],
        table_mode: str,
        tabs_names: Dict[str, str],
    ) -> List[SaleLineApi]:

        active_sale_lines = self._get_active_sale_lines(order_sales=order_sales)
        values_by_discount = self._get_discounts_from_sale_lines(sale_lines=active_sale_lines)

        sale_lines: List[SaleLineApi] = []
        fractions_dict: Dict[str, int] = {}
        for sale_line in active_sale_lines:

            custom_properties = sale_line.custom_properties
            self._set_sale_line_custom_property(
                custom_properties=custom_properties,
                key="orderId",
                value=str(order_id),
            )
            seat = self._get_seat(
                custom_properties=custom_properties,
                service_seats=service_seats,
                table_mode=table_mode,
                tabs_names=tabs_names,
            )

            self._set_fraction_dict(
                custom_properties=custom_properties,
                fractions_dict=fractions_dict,
                item_id=sale_line.item_id,
                line_number=sale_line.line_number,
                part_code=sale_line.part.part_code,
            )

            current_fraction_parent = fractions_dict.get(f"{sale_line.line_number}-{sale_line.item_id}", 0)
            if current_fraction_parent:
                sale_line.fraction_parent = current_fraction_parent

            sale_line.service_period = self._get_service_period(
                sales_line_list=active_sale_lines,
                sale_line=sale_line,
                properties=custom_properties,
            )

            if "benefitKey" in custom_properties:
                benefit_key = custom_properties.get("benefitKey")
                self._add_payments_from_benefit_to_seat(
                    seat=seat,
                    benefit_key=benefit_key,
                    payment_by_seat=payment_by_seat,
                    payment_by_benefit=payment_by_benefit,
                    used_benefit_keys=used_benefit_keys,
                )

            discount_value = self._get_discount_value(
                values_by_items=item_discount_values,
                values_by_discount=values_by_discount,
                tip_excluded_by_seat=tip_excluded_by_seat,
                discounts_tip_excluded=discounts_tip_excluded,
                line_id=sale_line.get_line_id(),
                item_price=sale_line.item_price,
                seat=seat
            )
            self._add_value_to_dict(
                key=seat,
                value=discount_value,
                dict_by_key=discount_by_seat,
            )

            sale_line.seat = seat
            sale_lines.append(sale_line)

        return sale_lines

    @staticmethod
    def _set_fraction_dict(
        custom_properties: Dict[str, str],
        fractions_dict: Dict[str, int],
        item_id: str,
        line_number: int,
        part_code: int,
    ) -> None:

        fraction = None
        if "fraction" in custom_properties:
            fraction = int(custom_properties.get("fraction") or 0)
            fraction = fraction if fraction > 1 else None

        if fraction:
            key = f"{line_number}-{item_id}.{part_code}"
            fractions_dict[key] = fraction

    @staticmethod
    def _get_seat(
        custom_properties: Dict[str, str],
        service_seats: int,
        table_mode: str,
        tabs_names: Dict[str, str],
    ) -> int:

        if table_mode == "tabs" and "tab" in custom_properties:
            return tabs_names.get(custom_properties.get("tab"), 0)

        if "seat" in custom_properties:
            seat = int(custom_properties.get("seat"))
            return 0 if seat > service_seats else seat

        return 0

    @staticmethod
    def _get_active_sale_lines(
        order_sales: List[SaleLineApi],
    ) -> List[SaleLineApi]:

        deleted_line_numbers = map(
            lambda x: x.line_number,
            filter(lambda x: x.level == 0 and x.quantity == 0, order_sales)
        )

        return list(filter(lambda x: x.line_number not in deleted_line_numbers, order_sales))

    def _get_discount_value(
        self,
        values_by_items: Dict[str, Dict[str, float]],
        values_by_discount: Dict[int, float],
        tip_excluded_by_seat: Dict[str, float],
        discounts_tip_excluded: List[int],
        line_id: str,
        item_price: float,
        seat: int,
    ) -> float:

        total_value = 0.0
        for discount_id in values_by_items.keys():
            for discount_line_number in values_by_items[discount_id]:
                if discount_line_number != line_id:
                    continue

                value = min(item_price, values_by_items[discount_id][discount_line_number])
                values_by_items[discount_id][discount_line_number] -= value
                values_by_discount[int(discount_id)] -= value
                total_value += value

                if int(discount_id) in discounts_tip_excluded:
                    self._add_value_to_dict(
                        key=seat,
                        value=value,
                        dict_by_key=tip_excluded_by_seat,
                    )

        return total_value

    def _add_payments_from_benefit_to_seat(
        self,
        seat: int,
        benefit_key: str,
        payment_by_seat: Dict[str, float],
        payment_by_benefit: Dict[str, float],
        used_benefit_keys: List[str],
    ) -> None:

        if benefit_key in used_benefit_keys or benefit_key not in payment_by_benefit:
            return

        self._add_value_to_dict(
            key=seat,
            value=payment_by_benefit[benefit_key],
            dict_by_key=payment_by_seat,
        )
        used_benefit_keys.append(benefit_key)

    def _get_service_period(
        self,
        sales_line_list: List[SaleLineApi],
        sale_line: SaleLineApi,
        properties: Dict[str, str],
    ) -> str:

        if sale_line.level == 0:
            return self._get_service_period_abbreviation(properties=properties)

        item_service_period = self._get_parent_service_period(
            sales_line_list=sales_line_list,
            line_number=int(sale_line.line_number),
        )

        return item_service_period

    def _get_parent_service_period(
        self,
        sales_line_list: List[SaleLineApi],
        line_number: int,
    ) -> Union[str, None]:

        for sale_line in sales_line_list:
            if sale_line.level != 0 or sale_line.line_number != line_number:
                continue

            return self._get_service_period_abbreviation(properties=sale_line.custom_properties)

    @staticmethod
    def _get_user_id_from_session_id(
        session_id: str,
    ) -> Optional[str]:

        for data in session_id.split(","):
            if data.startswith("user="):
                return data.split("=")[1]

        return None

    @staticmethod
    def _get_service_period_abbreviation(
        properties: Dict[str, str],
    ) -> Union[str, None]:

        if "servicePeriod" not in properties:
            return None

        props = json.loads(properties["servicePeriod"])
        return props.get("servicePeriodAbbr", None)

    def _get_discounts_from_sale_lines(
        self,
        sale_lines: List[SaleLineApi],
    ) -> Dict[int, float]:

        discount_values: Dict[int, float] = {}
        for sale_line in sale_lines:
            discounts_applied = sale_line.discounts_applied
            if discounts_applied is None:
                continue

            for discount in discounts_applied.split(","):
                discount_id, amount = discount.split(":")
                discount_id = int(discount_id)
                self._add_value_to_dict(
                    key=discount_id,
                    value=amount,
                    dict_by_key=discount_values,
                )

        return discount_values

    def _handle_custom_properties(
        self,
        custom_order_properties: List[CustomProperty],
        discounts_tip_excluded: List[int],
        order_tenders: List[Tender],
        payment_by_benefit_key: Dict[str, float],
        table: Table,
    ) -> Tuple[Dict[str, Dict[str, float]], str]:

        split_key = "1"
        split_key_found = False
        item_value_discounts: Dict[str, Dict[str, float]] = {}
        for custom_property in custom_order_properties:

            if not split_key_found and custom_property.key == "SPLIT_KEY":
                split_key = custom_property.value
                split_key_found = True

            if custom_property.key != "BENEFIT_LIST":
                continue

            benefit_list = json.loads(custom_property.value)
            self._parse_order_discounts(
                table=table,
                benefit_list=benefit_list,
            )
            item_value_discounts.update(benefit_list.get("discountItemValues", {}))
            discounts_tip_excluded.extend(benefit_list.get("discountIdTipExcluded", []))
            self._get_promotional_tenders(
                benefit_list=benefit_list,
                table=table,
                payment_by_benefit_key=payment_by_benefit_key,
                order_tenders=order_tenders,
            )

        return item_value_discounts, split_key

    def _get_promotional_tenders(
        self,
        benefit_list: Dict[str, str],
        table: Table,
        payment_by_benefit_key: Dict[str, float],
        order_tenders: List[Tender],
    ) -> None:

        benefits = json.loads(benefit_list.get("benefitList"))
        benefits_with_tender = [x for x in benefits if x.get("promotion_tender")]
        for benefit in benefits_with_tender:
            benefit_key = benefit.get("benefit_key")
            for tender in benefit.get("promotion_tender"):
                amount = float(tender.get("amount"))
                self._add_value_to_dict(
                    key=benefit_key,
                    value=amount,
                    dict_by_key=payment_by_benefit_key,
                )

        if not order_tenders or len(order_tenders) == 0:
            return

        tenders = {}
        for order_tender in order_tenders:
            self._add_value_to_dict(
                key=order_tender.tender_id,
                value=order_tender.amount,
                dict_by_key=tenders,
            )

        for benefit in benefits:
            for tender_id in benefit.get("order_tenders_ids"):
                amount = tenders.get(tender_id, 0)
                table.promotional_tenders += amount

    @staticmethod
    def _parse_order_discounts(
        table: Table,
        benefit_list: Dict,
    ) -> None:

        table_discount_groups: Dict[str, Dict[str, float]] = {}
        detailed_discounts = json.loads(benefit_list.get("benefitList", "[]"))
        for discount in detailed_discounts:
            group = discount.get("benefit", {}).get("group", "$OTHER_DISCOUNTS")
            discount_values = discount.get("discount_item_values", {})
            for discount_id, sale_lines in discount_values.items():
                for sale_line, discount_value in sale_lines.items():
                    if discount_value == 0:
                        continue

                    table_discount_groups.setdefault(group, {})
                    table_discount_groups[group].setdefault(sale_line, 0)
                    table_discount_groups[group][sale_line] += discount_value

        table.discount_groups = table_discount_groups

    @staticmethod
    def _add_value_to_dict(
        key: [int, str],
        value: float,
        dict_by_key: Dict[Union[str, int], float],
    ) -> None:

        if key in dict_by_key.keys():
            dict_by_key[key] += float(value)
        else:
            dict_by_key[key] = float(value)

    @staticmethod
    def _set_sale_line_custom_property(
        custom_properties: Dict[str, str],
        key: str,
        value: str,
    ) -> None:

        custom_properties.setdefault(key, value)



from ._DeliveryInfoDto import DeliveryInfoDto
from ._Tender import Tender
from ._CustomProperty import CustomProperty
from ._SaleLine import SaleLine
from ._Order import Order
from .order_xml_parser import parse_order



from typing import Dict

from messagebus import (
    MessageBus,
    SubscriptionInfo,
)
from ._EventPublishDetail import EventPublishDetail
from ._EventRepositoryFactory import EventRepositoryFactory
from ._NewEventListener import NewEventListener


class AsyncTargetReplySubscriber(NewEventListener):
    def __init__(
        self,
        message_bus: MessageBus,
        event_repository_factory: EventRepositoryFactory,
    ):
        self.message_bus = message_bus

        self.subscriptions: Dict[str, bool] = {}
        with event_repository_factory.build() as event_repository:
            pending_events = event_repository.get_pending()
            for pending_event in pending_events:
                if pending_event.async_target is None:
                    continue

                self._subscribe(reply_subject=pending_event.async_target.reply_subject)

    def new_event(self, publish_detail: EventPublishDetail):
        if not publish_detail.has_async_target():
            return

        self._subscribe(reply_subject=publish_detail.async_target.reply_subject)

    def _subscribe(self, reply_subject: str):
        if reply_subject in self.subscriptions:
            return

        self.message_bus.subscribe(
            [
                SubscriptionInfo(
                    subject=reply_subject,
                    subscription_type=SubscriptionInfo.TYPE_NON_REENTRANT,
                ),
            ]
        )
        self.subscriptions[reply_subject] = True



from typing import (
    Optional,
    List,
    Union,
    Dict,
)

from order_api.dto import SaleLine as SaleLineApi
from table_manager_api.dto import Table


class PreAccountSeatTotals(object):
    def __init__(
        self,
        sale_lines: List[SaleLineApi],
        table: Table,
        seat: Union[int, str],
        table_sub_total: float,
        table_tip: float,
        seat_quantities: int,
        table_payment: float,
    ) -> None:

        self.sale_lines = sale_lines
        self.table = table
        self.seat = seat

        self.seat_tip: Optional[int] = 0
        self.seat_total: Optional[int] = 0
        self.seat_payment: Optional[int] = 0
        self.total_seat_discount: Optional[int] = 0
        self.seat_total: Optional[int] = 0
        self.discounts_groups_sale_lines: Dict[str, float] = {}

        self._not_mapped_payment = table.promotional_tenders - sum(table.detailed_payments.values())
        self._sub_total_seats_amount = table.sub_total_amount - table_sub_total or 1
        self._table_tip_per_seat = table_tip / seat_quantities
        self._table_payment_per_seat = table_payment / seat_quantities

    def build_totals(
        self,
        seat_sub_total: float,
        seat_quantities: int,
        table_discount_per_seats: Dict[str, float],
        table_per_seat: float,
        tip_percentage: float,
    ) -> None:

        original_related_line_ids = self._get_original_related_line_ids(
            grouped_sale_lines=self.sale_lines,
        )

        self.total_seat_discount = self._get_total_seat_discount(
            table=self.table,
            original_related_line_ids=original_related_line_ids,
            seat_quantities=seat_quantities,
            table_discount_per_seats=table_discount_per_seats,
        )

        self.seat_payment = self._get_seat_payments_amount(
            table=self.table,
            seat=self.seat,
            seat_sub_total=seat_sub_total,
        )

        self.seat_tip = self._get_seat_tip_amount(
            table=self.table,
            seat=self.seat,
            seat_sub_total=seat_sub_total,
            tip_percentage=tip_percentage,
        )
        self.seat_total = seat_sub_total + table_per_seat - self.total_seat_discount + self.seat_tip - self.seat_payment

    def _get_total_seat_discount(
        self,
        table: Table,
        original_related_line_ids: List[str],
        seat_quantities: int,
        table_discount_per_seats: Dict[str, float],
    ) -> float:

        total_seat_discount = 0
        for key in table.discount_groups.keys():
            self.discounts_groups_sale_lines[key] = 0
            seat_discount = 0
            for discount_line_number in table.discount_groups[key]:
                discount_line_number = str(discount_line_number)
                if discount_line_number in original_related_line_ids:
                    seat_discount += table.discount_groups[key].get(discount_line_number, 0)

            table_discount_per_seat = table_discount_per_seats.get(key, 0) / seat_quantities
            seat_discount = round(seat_discount + table_discount_per_seat, 2)
            self.discounts_groups_sale_lines[key] = seat_discount
            total_seat_discount += seat_discount

        return total_seat_discount

    def _get_seat_payments_amount(
        self,
        table: Table,
        seat: Union[int, str],
        seat_sub_total: float,
    ) -> float:

        seat_payment = self._not_mapped_payment * (seat_sub_total / self._sub_total_seats_amount)
        seat_payment += table.detailed_payments[seat] if seat in table.detailed_payments.keys() else 0.0
        seat_payment = round(seat_payment + self._table_payment_per_seat, 2)
        return seat_payment

    def _get_seat_tip_amount(
        self,
        table: Table,
        seat: Union[int, str],
        seat_sub_total: float,
        tip_percentage: float,
    ) -> float:

        seat_tip_excluded = table.detailed_tip_excluded[seat] if seat in table.detailed_tip_excluded.keys() else 0.0
        seat_tip = (tip_percentage * (seat_sub_total - seat_tip_excluded)) / 100
        seat_tip = round(seat_tip + self._table_tip_per_seat, 2)
        return seat_tip

    @staticmethod
    def _get_original_related_line_ids(
        grouped_sale_lines: List[SaleLineApi],
    ) -> List[str]:
        original_related_line_ids: List[str] = []
        for grouped_sale_line in grouped_sale_lines:
            original_related_line_ids.extend(grouped_sale_line.original_related_line_ids or [])

        return original_related_line_ids
